import argparse

from . import pubCrawlLib

def download_pmid(pmid):
    meta = pubCrawlLib.getArticleMeta(pmid)
    pdf_file = pubCrawlLib.crawlOneDoc(meta, doc_type='pdf')
    return pdf_file

def download_pmid_program():
    parser = argparse.ArgumentParser(description='Process some integers.')
    parser.add_argument('pmid', type=str, help='The pubmed id to download')

    args = parser.parse_args()

    print(args)

    pdf_file = download_pmid(args.pmid)

    with open('{}.pdf'.format(args.pmid), 'wb') as file:
        file.write(pdf_file)




# generic functions for all pubtools, like logging, finding files, 
# ascii-conversion, section splitting etc

import os, logging, tempfile, sys, re, unicodedata, subprocess, time, types, traceback, \
    glob, operator, doctest, ftplib, random, shutil, atexit, pickle
from . import pubConf
from os.path import *
from distutils.spawn import find_executable

from time import gmtime, strftime

try:
    import leveldb
except:
    pass

# global var for the current headnode as specified on command line
forceHeadnode = None

# some data for countBadChars
all_chars = (chr(i) for i in range(0x110000))
specCodes = set(range(0,32))
goodCodes = set([7,9,10,11,12,13]) # BELL, TAB, LF, NL, FF (12), CR are not counted
badCharCodes = specCodes - goodCodes
control_chars = ''.join(map(chr, badCharCodes))
control_char_re = re.compile('[%s]' % re.escape(control_chars))

def getFastUniqueTempFname():
    " create unique tempdir on ramdisk, delete on exit "
    tempFname = tempfile.mktemp(dir=pubConf.getFastTempDir())
    maxCommon.delOnExit(tempFname)
    return tempFname

class Timeout(Exception):
    pass

def openKeyValDb(dbName, newDb=False, singleProcess=False, prefer=None):
    " factory function: returns the right db object given a filename "
    #return LevelDb(dbName, newDb=newDb)
    # possible other candidates: mdb, cdb, hamsterdb
    if prefer=="server":
        return pubKeyVal.RedisDb(dbName, newDb=newDb, singleProcess=singleProcess)
    else:
        return pubKeyVal.SqliteKvDb(dbName, newDb=newDb, singleProcess=singleProcess, tmpDir=pubConf.getFastTempDir())

def createDirRace(dirPath):
    """ create a directory on a cluster node, trying to fix race conditions """
    if not os.path.isdir(dirPath):
        time.sleep(random.randint(1,3)) # make sure that we are not all trying to create it at the same time
        if not os.path.isdir(dirPath):
            try:
                os.makedirs(dirPath)
            except OSError:
                logging.debug("Ignoring OSError, directory %s seems to exist already" % dirPath)

def getFromCache(fname):
    """ Given a network path, try to find a copy of this file on the local temp disk of a cluster node.
    Return the path on the local disk. If there is no copy yet, copy it over and return the path.
    """
    locCacheDir = join(pubConf.getTempDir(), "fileCache")
    createDirRace(locCacheDir)
    locPath = join(locCacheDir, basename(fname))
    logging.debug("Getting a local cache path for %s" % fname)
    if isfile(locPath):
        return locPath
    # it doesn't exist
    #fobj, locTmpName = makeTempFile(prefix=basename(fname), suffix=".tmp")
    # copy over to local temp name, takes a while
    locTmpName = tempfile.mktemp(prefix=basename(fname), suffix=".tmp")
    time.sleep(random.randint(1,3)+random.random()) # let's add some randomness
    logging.debug("Copying %s to %s" % (fname, locTmpName))
    shutil.copy(fname, locTmpName)
    # if another process copied it over by now: roll back
    if isfile(locPath):
        #fobj.close() # = delete
        os.remove(locTmpName)
        return locPath
    # move on local node to final name
    logging.debug("Moving %s to %s" % (locTmpName, locPath))
    shutil.move(locTmpName, locPath)
    return locPath

def runCommandTimeout(command, timeout=30, bufSize=128000, env=None, shell=True):
    """
    runs command, returns after timeout, kills subprocess if it takes longer than timeout seconds
    print runCommandTimeout(["ls", "-l"])
    print runCommandTimeout(["find", "/"], timeout=3) #should timeout

    returns stdout, stderr, ret
    """
    if type(command)==list:
        logging.log(5, "running command %s" % " ".join(command))
    else:
        logging.log(5, "running command %s" % command)
    if os.name=="nt":
        closeFds = False
    else:
        closeFds = True
    proc = subprocess.Popen(command, bufsize=bufSize, stdout=subprocess.PIPE, \
        stderr=subprocess.PIPE, shell=shell, close_fds=closeFds, env=env)
    poll_seconds = 1
    deadline = time.time()+timeout
    while time.time() < deadline and proc.poll() == None:
        time.sleep(poll_seconds)

    if proc.poll() == None:
        proc.terminate()
        time.sleep(1)
        proc.kill()
        logging.error("process %s timed out" % (str(command)))
        return "", "", 1

    stdout, stderr = proc.communicate()
    return stdout, stderr, proc.returncode

def findFiles(dir, extensions):
    """ find all files in or below dir with one of several extensions 
    extensions is a list, e.g. [".tab", ".txt"] or just a string like ".psl"

    returns typles: (relative directory to dir, full file path)
    """
    if isinstance(extensions, str): 
        extensions = [extensions]
    logging.debug("Reading filenames in %s with extensions %s" % (dir, str(extensions)))
    result = set()
    for dirPath, dirNames, fileNames in os.walk(dir):
        for fileName in fileNames:
            for ext in extensions:
                if fileName.endswith(ext):
                    #if (dirPath==dir): ?? Why this line ??
                        #continue
                    relDir = relpath(dirPath, dir)
                    fullPath = os.path.join(dirPath, fileName )
                    if os.path.getsize(fullPath)==0:
                        continue
                    result.add( (relDir, fullPath) )
    logging.info("Found %d files" % len(result))
    return result


def setupLoggingOptions(options):
    setupLogging("", options)
    
def verboseFunc(message):
    " we add this to logging "
    logging.log(5, message)

def addGeneralOptions(parser, noCluster=False, logDir=False, keepTemp=False):
    """ add the options that most cmd line programs accept to optparse parser object """
    parser.add_option("-d", "--debug", dest="debug", action="store_true", help="show debug messages")
    parser.add_option("-v", "--verbose", dest="verbose", action="store_true", help="show more debug messages")
    #parser.add_option("", "--quiet", dest="verbose", action="store_true", help="show more debug messages")
    if not noCluster:
        parser.add_option("-c", "--cluster", dest="cluster", action="store", help="override the default cluster head node from the config file, or 'localhost'")
    if logDir:
        parser.add_option("-l", "--logDir", dest="logDir", action="store", help="log messages to directory")
    if keepTemp:
        parser.add_option("", "--keepTemp", dest="keepTemp", action="store_true", help="keep temporary files, for debugging")
    return parser

def logToFile(logFileName):
    " add a log handler that write to a file and return the handler "
    rootLog = logging.getLogger('')
    fh = logging.FileHandler(logFileName)
    fh.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    fh.setFormatter(formatter)
    rootLog.addHandler(fh)
    return fh

debugMode=False

def setupLogging(progName, options, parser=None, logFileName=None, \
        debug=False, fileLevel=logging.DEBUG, minimumLog=False, fileMode="w"):
    """ direct logging to a file and also to stdout, depending on options (debug, verbose, jobId, etc) """
    assert(progName!=None)
    global debugMode
    logging.addLevelName(5,"VERBOSE")

    if options!=None and "cluster" in options.__dict__ and options.cluster!=None:
        global forceHeadnode
        # for makeClusterRunner
        forceHeadnode = options.cluster

    if options!=None and "keepTemp" in options.__dict__ and options.keepTemp==True:
        maxCommon.keepTemp=True

    if options==None:
        stdoutLevel=logging.DEBUG

    elif "verbose" in options.__dict__ and options.verbose:
        stdoutLevel=3
        fileLevel = 3
        debugMode = True

    elif options.debug or debug:
        stdoutLevel=logging.DEBUG
        debugMode = True

    elif minimumLog:
        stdoutLevel=logging.ERROR

    elif "logDir" in options.__dict__ and options.logDir:
        stdoutLevel=logging.CRITICAL
        fileLevel = logging.INFO
        debugMode = True

        timeStr = strftime("_%Y-%m-%d_%H%M%S", gmtime())
        logFileName = join(options.logDir, progName+timeStr)
        logging.warn("Writing message log to %s" % logFileName)

    else:
        stdoutLevel=logging.INFO

    rootLog = logging.getLogger('')
    rootLog.setLevel(fileLevel)
    logging.root.handlers = []

    logging.verbose = verboseFunc
    # setup file logger
    if logFileName != None and logFileName!="":
        fh = logging.FileHandler(logFileName)
        fh.setLevel(logging.DEBUG)
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        fh.setFormatter(formatter)
        rootLog.addHandler(fh)

    #if logFileName != None:
        #logging.basicConfig(level=fileLevel,
                            #format='%(asctime)s %(name)-12s %(levelname)-8s %(message)s',
                            #datefmt='%m-%d %H:%M',
                            #filename= logFileName,
                            #filemode=fileMode, stream=None)
    # define a handler which writes messages to sys.stderr
    console = logging.StreamHandler()
    # set a format which is simpler for console use
    formatter = logging.Formatter('%(levelname)-8s-%(message)s')
    # tell the handler to use this format
    console.setFormatter(formatter)
    console.setLevel(stdoutLevel)
    # make sure that the root logger gets verbose messages 
    logging.getLogger('').setLevel(min(stdoutLevel, fileLevel))
    # add the handler to the root logger
    rootLog.addHandler(console)

def printOut(stdout, stderr):
    """ send two strings to logging """
    if len(stdout)!=0:
        logging.debug("stdout: %s" % stdout)
    if len(stderr)!=0:
        if len(stderr) > 1000 or "a4 is redefined" in stderr:
            logging.info("not showing stderr, too big or containes annoying message")
        else:
            logging.info("Size of stderr: %d" % len(stderr))
            logging.info("stderr: %s" % stderr)

def runConverter(cmdLine, fileContent, fileExt, tempDir):
    """ create local temp for in and output files, write data to infile, run
    command. file can be supplied as a str in fileContent["content"] or
    alternatively as a pathname via 'locFname' """
    # create temp file
    fd, inFname = tempfile.mkstemp(suffix="."+fileExt, dir=tempDir, prefix="pubConvPmc.in.")
    logging.debug("Created %s" % inFname)
    maxCommon.delOnExit(inFname)

    inFile = os.fdopen(fd, "wb")
    inFile.write(fileContent)
    inFile.close()

    logging.debug("Created in temp file %s" % (inFname))

    fd, outFname = tempfile.mkstemp(suffix=".txt", dir=tempDir, prefix="pubConvPmc.out.")
    maxCommon.delOnExit(outFname)
    os.close(fd)
    logging.debug("Created out temp file %s" % (outFname))

    # allow %(name)s syntax in cmdLine string to use variables from pubConf
    cmdLine = cmdLine % pubConf.__dict__
    # build cmd line and run
    cmdLine = cmdLine.replace("$in", inFname)
    cmdLine = cmdLine.replace("$out", outFname)
    logging.debug("running "+cmdLine)
    skipFile=False
    stdout, stderr, ret = runCommandTimeout(cmdLine, bufSize=10000000, timeout=30)

    asciiData = None

    if ret==2 and "docx2txt" not in cmdLine: # docx2text returns exit code 2 in some cases
        logging.error("stopped on errno 2: looks like you pressed ctrl-c")
        sys.exit(2)

    if ret!=0:
        logging.error("error %d occured while executing %s" % (ret, cmdLine))
        printOut(stdout, stderr)
        skipFile=True
    if os.path.getsize(outFname)==0:
        logging.error("zero file size of output file after command %s" % (cmdLine))
        printOut(stdout, stderr)
        skipFile=True

    if not skipFile:
        asciiData = open(outFname).read()

    logging.debug("Removing %s" % inFname)
    os.remove(inFname)
    os.remove(outFname)

    asciiData = forceToUnicode(asciiData)

    if skipFile:
        return None
    else:
        return asciiData

# http://stackoverflow.com/questions/92438/stripping-non-printable-characters-from-a-string-in-python
def countBadChars(string):
    """ count special chars in string, but not tab, cr, nl, etc """
    #print "COUNT BAD CHARS", repr(string)
    badCharCount = len(list(control_char_re.finditer(string)))
    #print "BAD CHARS", badCharCount
    return badCharCount

# http://stackoverflow.com/questions/92438/stripping-non-printable-characters-from-a-string-in-python
def removeBadChars(string):
    """ replace bad chars with spaces """
    return control_char_re.sub(" ", string)

def getFileExt(fileData, locFname, mimeType):
    " try to determine best externsion for a file given its fileData dict "
    url = fileData["url"]
    fileExt = None

    logging.debug("trying mime type to get extension")
    # get the mimeType, either local or from dict
    if mimeType==None and "mimeType" in fileData and fileData["mimeType"]!=None:
        mimeType = fileData["mimeType"]
        logging.debug("mime type is %s" % mimeType)

    # get extensions from mime TYpe
    if mimeType!=None:
        fileExt = pubConf.MIMEMAP.get(mimeType, None)
        logging.debug("File extension based on mime type %s" % mimeType)

    # this used to be different - did it break something?
    if fileExt==None:
        if locFname:
            logging.debug("File extension taken from local file %s" % locFname)
            filePath = locFname
        else:
            logging.debug("File extension taken from url %s" % url)
            filePath = url

        fileExt = os.path.splitext(filePath)[1].lower().strip(".")

    logging.debug("File extension determined as  %s" % fileExt)
    return fileExt

def toAscii(fileData, mimeType=None, \
        maxBinFileSize=pubConf.maxBinFileSize, maxTxtFileSize=pubConf.maxTxtFileSize, \
        minTxtFileSize=pubConf.minTxtFileSize):
    """ pick out the content from the fileData dictionary, 
    write it to a local file in tempDir and convert it to 
    ASCII format. Put output back into the content field.

    mimeType will be used if specified, otherwise try to guess
    converter based on url file extension

    returns fileData if successful, otherwise None
    returns only unicode strings (despite the name)
    """
    converters = pubConf.getConverters()
    tempDir = pubConf.getTempDir()

    fileContent = fileData["content"]
    fileSize = len(fileContent)

    if "locFname" in fileData:
        locFname=fileData["locFname"]
        fileDebugDesc = fileData["externalId"]+":"+locFname
    else:
        locFname = None
        fileDebugDesc = ",".join([fileData["url"],fileData["desc"],
            fileData["fileId"],fileData["articleId"]])

    if fileSize > maxBinFileSize:
        logging.warn("binary file size before conversion %d > %d, skipping file %s" % \
            (len(fileContent), maxBinFileSize, fileDebugDesc))
        return None

    fileExt = getFileExt(fileData, locFname, mimeType)

    if fileExt not in converters:
        logging.debug("Could not convert file %s, no converter for extension %s" % \
            (fileDebugDesc, fileExt))
        return None
    cmdLine = converters[fileExt]

    if cmdLine=="COPY":
        # fileData["content"] already contains ASCII text
        pass
        
    elif cmdLine=="XMLTEXT" or cmdLine=="NXMLTEXT":
        logging.debug("stripping XML tags")
        if cmdLine=="NXMLTEXT":
            asciiData = pubXml.stripXmlTags(fileContent, isNxmlFormat=True)
        else:
            asciiData = pubXml.stripXmlTags(fileContent)

        if asciiData==None:
            logging.debug("Could not convert xml to ascii, file %s" % fileData["url"])
            return None
        fileData["content"]=asciiData

    else:
        asciiData = runConverter(cmdLine, fileContent, fileExt, tempDir)

        # try to detect corrupted pdf2text output and run second converter
        if fileExt=="pdf" and \
            ((asciiData==None or len(asciiData)<minTxtFileSize) or countBadChars(asciiData)>=10):
            logging.debug("No data or too many non printable characters in PDF, trying alternative program")
            cmdLine = converters["pdf2"]
            asciiData = runConverter(cmdLine, fileContent, fileExt, tempDir)

        if asciiData==None:
            logging.info("conversion failed for %s" % fileDebugDesc)
            return None
        else:
            fileData["content"]=removeBadChars(asciiData)

    fileData = dictToUnicode(fileData)

    if len(fileData["content"]) > maxTxtFileSize:
        logging.info("ascii file size after conversion too big, ignoring file %s" % fileDebugDesc)
        return None

    if len(fileData["content"]) < minTxtFileSize:
        logging.debug("ascii file size only %d bytes < %d, ignoring %s" % \
            (len(fileData["content"]), minTxtFileSize, fileDebugDesc))
        return None

    #charSet = set(fileData["content"])
    #if len(charSet) < 10:
        #logging.warn("too few characters in ASCII output: %s" % charSet)
        #return None

    return fileData

def toAsciiEscape(fileData, mimeType=None, maxBinFileSize=pubConf.maxBinFileSize, maxTxtFileSize=pubConf.maxBinFileSize, minTxtFileSize=pubConf.minTxtFileSize):
    """ convert to ascii, escape special characters 
        returns a fileData dict
    """
    fileData = toAscii(fileData, mimeType=mimeType,\
            maxBinFileSize=maxBinFileSize, maxTxtFileSize=maxTxtFileSize, minTxtFileSize=minTxtFileSize)
    fileData = pubStore.dictToUtf8Escape(fileData)
    return fileData

def stringListToDict(paramList):
    """ splits all strings at '=' and returns a dict from these 
    input: ["a=3", "b=bla"] 
    output: {"a":"3","b":"bla"}
    """
    paramItems = [string.split("=", 1) for string in paramList]
    paramDict = {}
    for key, val in paramItems:
        paramDict[key] = val
    return paramDict

def readArticleChunkAssignment(inDir, updateIds):
    "read the assignment of articleId -> chunkId from text directory"

    if updateIds==None:
        inFiles = glob.glob(os.path.join(inDir, "*_index.tab"))
    else:
        inFiles = []
        for updateId in updateIds:
            updateId = str(updateId)
            indexFname = "%s_index.tab" % updateId
            if isfile(indexFname):
                inFiles.append(os.path.join(inDir, indexFname))

    if len(inFiles)==0:
        logging.warn("No article chunk assignment")
        return None

    logging.debug("Input files for article -> chunk assignment: %s" % inFiles)

    articleChunks = {}
    for inFile in inFiles:
        logging.info("Parsing %s" % inFile)
        for row in maxCommon.iterTsvRows(inFile):
            chunkId = int(row.chunkId.split("_")[1])
            articleChunks[int(row.articleId)] = int(chunkId)
    return articleChunks

def forceToUnicode(text):
    " force to unicode string: try utf8 first, then latin1 "
    if text==None:
        return None
    if type(text)==str:
        #logging.debug("text is unicode")
        return text
    try:
        text = text.decode("utf8")
    except Exception as err:
        logging.debug("Could not convert to unicode using utf8, problem %s, traceback to stdout" % (err))
        #traceback.print_exception(*sys.exc_info())
        try:
            text = text.decode("latin1")
            logging.debug("Converted using latin1")
        except Exception as err:
            logging.debug("Could not convert to unicode using latin1, problem %s" % err)
            try:
                text = text.decode("cp1252")
            except Exception as err:
                logging.debug("Could not convert to unicode using cp1252, problem %s, traceback to stdout" % (err))
            pass
    return text

def dictToUnicode(dict):
    " forcing all values of dict to unicode strings "
    result = {}
    for key, val in dict.items():
        result[key] = forceToUnicode(val)
    return result

def lftpGet(remoteUrl, locDir, fileNames, connCount):
    " use lftp to download files in parallel "
    locDir = os.path.abspath(locDir)
    scriptPath = join(locDir, "lftp.cmd")
    logging.debug("Writing filenames to %s" % scriptPath)
    lFile = open(scriptPath, "w")
    lFile.write("set net:socket-buffer 32000000\n")
    lFile.write("set cmd:parallel %d\n" % int(connCount))
    lFile.write("set xfer:use-temp-file yes\n")  # atomic download
    lFile.write("set xfer:clobber yes\n")
    lFile.write("open %s\n" % remoteUrl)
    lFile.write("set xfer:log true\n")
    lFile.write("lcd %s\n" % locDir)
    pm = maxCommon.ProgressMeter(len(fileNames))
    existDirs = set()
    locNames = []
    for f in fileNames:
        locName = join(locDir, f)
        # make sure that target dir exists
        locFileDir = dirname(locName)
        if locFileDir not in existDirs and not isdir(locFileDir):
            logging.info("Creating directory %s" % locFileDir)
            os.makedirs(locFileDir)
        existDirs.add(locFileDir)

        logging.debug("filename %s" % locName)
        if isfile(locName):
            logging.debug("Already exists: %s, skipping" % locName)
        else:
            lFile.write("get %s -o %s\n" % (f, locName))
            locNames.append(locName)
        pm.taskCompleted()
    lFile.close()

    if find_executable("lftp") is None:
        raise Exception("the command lftp is not in your PATH. Install it wit apt-get install lftp or yum install lftp")

    cmd = ["lftp", "-f", scriptPath]
    logging.debug("Launching lftp for download, cmd %s" % " ".join(cmd))
    ret = subprocess.call(cmd)

    if ret!=0:
        logging.error("error during transfer")
        sys.exit(1)

    logging.debug("Updating downloads.log file in %s" % locDir)
    for f in fileNames:
        appendLog(locDir, "add", f)
    logging.info("Downloaded %d files: %s" % (len(locNames), str(",".join(locNames))))

def appendLog(outDir, change, fname):
    logPath = join(outDir, "downloads.log")
    if not isfile(logPath):
        logFile = open(logPath, "w")
        logFile.write("date\tchange\tfiles\n")
    else:
        logFile = open(logPath, "a")

    timeStr = time.strftime("%x %X")
    row = [timeStr, change, fname]
    logFile.write("\t".join(row))
    logFile.write("\n")
    logFile.close()

def getFtpDir(server, path):
    " return list of files on ftp server "
    logging.info("Getting FTP directory server %s, %s" % (server, path))
    ftp = ftplib.FTP(server, user="anonymous", passwd=pubConf.email, timeout=60)
    ftp.cwd(path)
    dirLines = ftp.nlst() 
    #fnames = [split(line)[0] for line in dirLines]
    logging.debug("Found files %s" % dirLines)
    return dirLines


def resolveDatasetDesc(descs):
    " resolve a comma-sep list of dataset identifiers like pmc or elsevier to a list of directories "
    dirs = []
    for desc in descs.split(','):
        descDir = pubConf.resolveTextDir(desc)
        if descDir==None:
            raise Exception("Unknown dataset: %s" % desc)
        dirs.append(descDir)
    return dirs

def splitAnnotId(annotId):
    """
    split the 64bit-annotId into packs of 10/3/5 digits and return all
    >>> splitAnnotId(200616640112350013)
    (2006166401, 123, 50013)
    """
    fileDigits = pubConf.FILEDIGITS
    annotDigits = pubConf.ANNOTDIGITS
    articleDigits = pubConf.ARTICLEDIGITS

    annotIdInt = int(annotId)
    articleId  = annotIdInt / 10**(fileDigits+annotDigits)
    fileAnnotId= annotIdInt % 10**(fileDigits+annotDigits)
    fileId     = fileAnnotId / 10**(annotDigits)
    annotId    = fileAnnotId % 10**(annotDigits)
    return articleId, fileId, annotId

def splitAnnotIdString(annotIdString):
    """ split annot as a string into three parts 
    >>> splitAnnotId("200616640112350013")
    (2006166401, 123, 50013)
    """
    fileDigits = pubConf.FILEDIGITS
    annotDigits = pubConf.ANNOTDIGITS
    articleDigits = pubConf.ARTICLEDIGITS

    articleId = annotIdString[:articleDigits]
    fileId = annotIdString[articleDigits:articleDigits+fileDigits]
    annotId = annotIdString[articleDigits+fileDigits:]
    return articleId, fileId, annotId

def makeTempDir(prefix, tmpDir=None):
    """ create unique local temp subdir in pubtools temp dir.
        the pubtools temp dir is usually located on a local disk or ramdisk.
    """
    if tmpDir==None:
        tmpDir=pubConf.getTempDir()
    dirName = tempfile.mktemp(dir=tmpDir, prefix=prefix+".")
    if not isdir(dirName):
        os.makedirs(dirName)
    logging.debug("Created temporary dir %s" % dirName)
    return dirName

def makeTempFile(prefix, suffix=".psl"):
    """ create tempfile in pubtools tempdir dir with given prefix. 
    Return tuple (file object , name).
    Tempfile will auto-delete when file object is destructed, unless debug mode is set. 
    """
    tmpDir=pubConf.getTempDir()
    if pubConf.debug:
        tfname = tempfile.mktemp(dir=tmpDir, prefix=prefix+".", suffix=suffix)
        #tfname = join(tmpDir, prefix+suffix)
        tf = open(tfname, "w")
        logging.debug("Created tempfile %s, debug-mode: no auto-deletion" % tfname)
    else:
        tf = tempfile.NamedTemporaryFile(dir=tmpDir, prefix=prefix+".", mode="w", suffix=suffix)
    return tf, tf.name

def concatDelLogs(inDir, outDir, outFname):
    " concat all log files to outFname in outDir and delete them "
    outPath = join(outDir, outFname)
    inMask = join(inDir, "*_*.log")
    logFnames = glob.glob(inMask)
    logging.info("Concatting %d logfiles from %s to %s" % (len(logFnames), inMask, outPath))
    delFnames = []

    if not isfile(outPath):
        ofh = open(outPath, "w")
        for inFname in logFnames:
            ofh.write("---- LOGFILE %s ------\n" % inFname)
            ofh.write(open(inFname).read())
            ofh.write("\n")
            delFnames.append(inFname)
        ofh.close()

    for fname in delFnames:
        os.remove(fname)

def concatIdentifiers(inDir, outDir, outFname):
    " concat all identifiers of *_ids.tab files in inDir to outFname, append if exists "
    outPath = join(outDir, outFname)
    inMask = join(inDir, "*_ids.tab")
    idFnames = glob.glob(inMask)
    logging.debug("Concatting exernalIds from %s to %s" % (inMask, outPath))
    extIds = []
    for inFname in idFnames:
        if os.path.getsize(inFname)==0:
            logging.warn("file %s has zero size")
            continue
        for row in maxCommon.iterTsvRows(inFname):
            extIds.append(row.externalId)

    if isfile(outPath):
        ofh = open(outPath, "a")
    else:
        ofh = open(outPath, "w")
        ofh.write("#externalId\n")

    for extId in extIds:
        ofh.write("%s\n" % extId)
    ofh.close()

    return outPath
    
def parseDoneIds(fname):
    " parse all already converted identifiers from inDir "
    print(fname)
    doneIds = set()
    if os.path.getsize(fname)==0:
        return doneIds

    for row in maxCommon.iterTsvRows(fname):
        doneIds.add(row.externalId)
    logging.info("Found %d identifiers of already parsed articles" % len(doneIds))
    return doneIds
            
def concatDelIdFiles(inDir, outDir, outFname):
    """ concat all id files in outDir, write to outFname, delete all id files when finished 
    """
    outPath = join(outDir, outFname)
    inMask = join(inDir, "*ids.tab")
    idFnames = glob.glob(inMask)
    logging.debug("Concatting %s to %s" % (inMask, outPath))
    maxTables.concatHeaderTabFiles(idFnames, outPath)
    maxCommon.deleteFiles(idFnames)
    return outPath

lockFnames = []

def setLockFile(outDir, lockName):
    """ 
    create lock file. die if already exists
    """
    global lockFnames
    lockFname = join(outDir, lockName+".lock")
    if isfile(lockFname):
        raise Exception("The lockfile %s exists. Make sure there is no process already running, "
        "then delete the lock file." % lockFname)
    open(lockFname, "w")
    lockFnames.append(lockFname)
    atexit.register(removeLockFiles)

def removeLockFiles():
    """ 
    remove all lock files. Die if not exists.
    """
    for lockFname in lockFnames:
        if not isfile(lockFname):
            #raise Exception("%s not found." % lockFname)
            logging.debug("lockfile %s does not exist, skipping" % lockFname)
            continue
        logging.debug("Removing %s" % lockFname)
        os.remove(lockFname)

def setInOutDirs(useDefault, args, pubName):
    """ get in and output dir either from args or create default values,
    depending on useDefault """

    if useDefault:
        inDir, outDir = join(pubConf.extDir, pubName), join(pubConf.textDir, pubName)
    else:
        inDir, outDir = args[:2]
    return inDir, outDir

def makeBuildDir(outDir, mustExist=False):
    ' create a dir outDir/build and error abort if it already exists '
    outDir     = join(outDir, "build")
    if not mustExist and isdir(outDir):
        raise Exception("Directory %s already exists. Looks like a previous conversion run "
            "crashed or is ongoing. Delete the directory and re-run if needed or complete the batch "
            "and use pubConvXXX with the --finish option" % outDir)
    if mustExist and not isdir(outDir):
        raise Exception("Directory %s does not exist." % outDir)

    if not mustExist:
        os.mkdir(outDir)
    return outDir

def saveUpdateInfo(buildDir, updateId, lastArticleId, newIds):
    " save often-needed data to a pickle file in buildDir "
    updInfo = {'updateId' : updateId, 'lastArticleId' : lastArticleId, 'newIds' : newIds}
    updInfoFh = open(join(buildDir, "updateInfo.pickle"), "w")
    logging.info("Saving update info to %s" % updInfoFh.name)
    pickle.dump(updInfo, updInfoFh)

def loadUpdateInfo(buildDir):
    " save often-needed data to a pickle file in buildDir "
    updFname = join(buildDir, "updateInfo.pickle")
    logging.info("loading update info from %s" % updFname)
    updInfoFh = open(updFname)
    updInfo = pickle.load(updInfoFh)
    return updInfo["updateId"], updInfo['lastArticleId'], updInfo['newIds']

def removeUpdateInfo(buildDir):
    updFname = join(buildDir, "updateInfo.pickle")
    logging.debug("Removing %s" % updFname)
    os.remove(updFname)

if __name__=="__main__":
    setupLoggingOptions(None)
    #doctest.testmod()
    test_sectioning()
    #print sectionRes["results"].search("bla\nResults").start()
    #print list(re.compile(r"^Results").findall("bla\nResults"))



# Routines for reading/writing tables from/to textfiles or databases

import sys, textwrap, operator, types, logging, re, os, collections, codecs, time, gzip, shutil

# jython doesn't have sqlite3
try:
    import sqlite3
except ImportError:
    pass

from types import *

from . import maxCommon, pubGeneric

# Routines for handling fasta sequences and tab sep files
try:
    import MySQLdb
except:
    #logging.warn("MySQLdb is not installed on this machine, tables cannot be read from mysql")
    pass

def openFileRead(fname):
    return openFile(fname)

def openFile(fname, mode="r"):
    """ open and return filehandle, open stdin if fname=="stdin", do nothing if none """
    if fname=="stdin":
        return sys.stdin
    elif fname.endswith(".gz"):
        return gzip.open(fname)
    elif fname=="stdout":
        return sys.stdout
    elif fname=="none" or fname==None:
        return None

    if mode=="r":
        if not os.path.isfile(fname):
            logging.error("File %s does not exist" % fname)
        else:
            if os.path.getsize(fname)==0:
                logging.warn("File %s is empty" % fname)

    return open(fname, mode)

###
class BlockReader:
    """ a parser for sorted, tables: Splits a table 'blocks' where values in
    one field are identical. Returns one block at a time. As such, it can
    process GB of tables, without having to load them into memory. 
    
    if mustBeSorted==True: Tables MUST BE sorted with sort: remember to use the
    -n option, do not pipe into sort
    to save memory. """

    def __init__(self, fileObj, idField, mustBeSorted=True):
        """ initialize blockReader to read from file fname, the idField is the field which contains the id that the file is sorted on """
        #self.fh = open(fname, "r")
        if type(fileObj)==FileType:
            self.fh = fileObj
        else:
            self.fh = open(fileObj)
        self.block = []
        self.lineCount=0
        self.blockIds=set()
        self.lastId = None
        self.idField = int(idField)
        self.mustBeSorted=mustBeSorted
        self.headers = self.fh.readline().strip("\n").strip("#").split("\t")

    def readNext(self):
        """ read next block and return as list of tuples """
        blockId=None
        for l in self.fh:
            self.lineCount+=1
            tuple = l.strip("\n").split("\t")
            blockId = tuple[self.idField]
            if self.mustBeSorted:
                assert(self.lastId==None or self.lastId=='None' or (int(blockId) >= int(self.lastId))) # infile for this script MUST BE sorted with sort (remember to use the -n option, do not pipe into sort) 
            
            if blockId==self.lastId or self.lastId==None:
                # no change or first block: just add line
                self.block.append(tuple)
                self.lastId=blockId
            else:
                # change of blockId: clear block and update blockId
                self.blockIds.add(blockId)
                oldBlock=self.block
                oldId=self.lastId

                self.block=[tuple]
                self.lastId = blockId
                yield oldId, oldBlock

        # for last line
        if blockId!=None:
            self.blockIds.add(blockId)
        yield self.lastId, self.block


###
class TableParser:
    """ Class to read tables from (tab-sep) textfiles. Fieldnames can be read from textfile headers or are provided, fields are converted to specified data types (int/string/float). 

    >>> docStr="#id\\tfirstname\\tlastname\\n#test\\n1\\tmax\\thaussler\\n"
    >>> import StringIO
    >>> f = StringIO.StringIO(docStr)
    >>> tp = TableParser(fileObj=f)
    >>> print list(tp.lines())
    [tuple(id='1', firstname='max', lastname='haussler')]
    >>> tp2 = TableParser(None, headers=['field1', 'field2'])
    >>> print tp2.parseTuple(['test1', 'test2'])
    tuple(field1='test1', field2='test2')
    """

    def __init__(self, fileObj, headers=None, fileType=None, types=None, colCount=None, encoding="utf8"):
        """ 
        Parse headers from file (read only first line from file)
        or parse headers from headers parameter
        or use predefined headers according to fileType
        fileType can be: numColumns, blastm8, psl, blastConvert, bed3, bed4, sam 

        specify numColumns if file doesn't have a header line
        """

        self.types=types

        if isinstance(fileObj, str):
            fileObj = codecs.open(fileObj, encoding=encoding)

        if fileObj:
            self.fileObj = fileObj

        self.commentChar="#"
        self.line1=None

        if fileType==None and headers==None:
            # parse headers from file and set types 
            # all to String
            line1 = fileObj.readline().strip("\n")
            line1 = line1.strip(self.commentChar)
            self.headers = line1.split("\t")
        elif headers:
            self.headers = headers
        else:
            # predefined file formats, set your editor to nowrap lines to read them better
            if fileType=="numbered" or fileType=="numColumns":
                if colCount==None:
                    self.line1 = fileObj.readline()
                    colCount = len(self.line1.split("\t"))
                self.headers = ["col"+str(i) for i in range(0, int(colCount))]
                self.types   = [StringType] * len(self.headers)
            elif fileType=="psl":
                self.headers = ["score", "misMatches", "repMatches", "nCount", "qNumInsert", "qBaseInsert", "tNumInsert", "tBaseInsert", "strand",    "qName",    "qSize", "qStart", "qEnd", "tName",    "tSize", "tStart", "tEnd", "blockCount", "blockSizes", "qStarts", "tStarts"]
                self.types =   [IntType, IntType,       IntType,      IntType,  IntType,      IntType,       IntType,       IntType,      StringType, StringType, IntType, IntType,  IntType, StringType, IntType, IntType,  IntType, IntType ,   StringType,      StringType,   StringType]
            elif fileType=="blastm8":
                self.headers = ["qName",    "tName",    "percIdentity", "alnLength", "misMatches", "gapOpenCount", "qStart", "qEnd", "tStart", "tEnd",  "eVal",    "score"]
                self.types =   [StringType, StringType, FloatType,      IntType,      IntType,      IntType,        IntType, IntType, IntType, IntType, FloatType, IntType,]
            elif fileType=="intmap":
                self.headers = ["int", "string"]
                self.types = [IntType, StringType]
            elif fileType=="blastConvert":
                self.headers = ["pmcId", "genomeId", "seqId", "chrom", "tStart", "tEnd", "score"]
                self.types =   [IntType, IntType,    IntType, StringType, IntType, IntType, FloatType]
            elif fileType=="bed4":
                self.headers = ["chrom", "start", "end", "name"]
                self.types =   [StringType, IntType, IntType, StringType]
            elif fileType=="bed3":
                self.headers = ["chrom", "start", "end"]
                self.types =   [StringType, IntType, IntType]
            elif fileType == "sam":
                self.headers = ["qname"    , "flag"  , "rname"    , "pos"   , "mapq"  , "cigar"    , "nrnm"     , "mpos"  , "isize" , "seq"      , "qual"     , "tags"]
                self.types   = [StringType , IntType , StringType , IntType , IntType , StringType , StringType , IntType , IntType , StringType , StringType , StringType]
                self.commentChar="@"
                
            else:
                logging.error("maxTables.py: illegal fileType\n")
                sys.exit(1)

        if headers:
            self.headers = headers
        if not self.types:
            self.types = [StringType] * len(self.headers)

        #logging.debug("Headers are: %s" % str(self.headers))
        self.Record = collections.namedtuple("tuple", self.headers) # this is a backport from python2.6

    def parseTuple(self, tuple):
        """
        >>> TableParser(None, fileType="intmap").parseTuple(["1", "hallo"])
        tuple(int=1, string='hallo')

        """
        # convert fields to correct data type
        if self.types: 
            tuple = [f(x) for f, x in zip(self.types, tuple)]
        # convert tuple to record with named fields
        return self.Record(*tuple)    

    def parseBlock(self, block):
        """ convert list of tuples to list of records """
        newBlock= []
        for tuple in block:
            newBlock.append(self.parseTuple(tuple))
        return newBlock

    def parseLine(self, line):
        tuple = line.strip("\n").split("\t")    
        return self.parseTuple(tuple)
    
    def generateRows(self):
        """ just another name for lines() """
        return self.lines()

    def rows(self):
        """ and another name for lines() """
        return self.lines()

    def lines(self):
        """ Generator: return next tuple, will skip over empty lines and comments (=lines that start with #). """
        while True:
            line = '\n'
            if self.line1!=None:
                line = self.line1
                self.line1=None
                yield self.parseLine(line)

            while line=='\n' or line.startswith(self.commentChar) or line =="\r" or line =='\r\n':
                line = self.fileObj.readline()
            if line=='':
                return 
            yield self.parseLine(line)

    def column(self, columnName, dataType=bytes):
        """ Generator: return values of a given column (name) """
        for row in self.lines():
            yield dataType(row._asdict()[columnName])

    def columns(self, columnNames):
        """ Generator: return values of given columns (name) """
        for row in self.lines():
            result = []
            rowDict = row._asdict()
            for columnName in columnNames:
                result.append(rowDict[columnName])
            yield result
# ---- parse rows from mysql table ------

def hgSqlConnect(db, config=None, **kwargs):
    " connect using information parsed from ~/.hg.conf "
    if config==None:
        config = maxCommon.parseConfig("~/.hg.conf")
    conn = MySQLdb.connect(host=config["db.host"], user=config["db.user"], \
        passwd=config["db.password"], db=db, **kwargs)
    return conn

def sqlConnection(connString=None, **kwargs):
    """ connectionString format: hostWithPort,user,password,db 
    > conn = sqlConnection(connString="localhost,max,test,temp")
    """
    if connString:
        host,user,passwd,db = connString.split(",")

        port = 0
        fs = host.split(":")
        host = fs[0]
        if len(fs)>1:
            port=int(fs[1])

        db = MySQLdb.connect(host, user, passwd, db, port=port)
    else:
        try:
            db = MySQLdb.connect(**kwargs)
        except:
            kwargs["read_default_file"]="~/.my.cnf"
            db = MySQLdb.connect(**kwargs)
    return db

def sqlGetRows(conn, sqlString, *args):
    """ execute sqlString with placeholders %s replaced by args
    caveat: if there is only one arg, you need to send it as (arg,) ! see mysqldb docs
    returns a dictionary fieldname -> value
    """
    cursor = conn.cursor(MySQLdb.cursors.DictCursor)
    cursor.execute(sqlString, *args)
    rows = cursor.fetchall()

    # check if any column is an array (bug in mysqldb 1.2.1 but not in 1.2.2)
    #if len(rows)>0:
        #arrayType = type(array.array('c', "a"))
        #needConvert = False
        #row1 = rows[0]
        #for d in row1:
            #if type(d)==arrayType:
                #needConvert=True
                #break
        #if needConvert:
            #newRows = []
            #for row in rows:
                #newCols = []
                #for col in row:
                    #if type(col)==arrayType:
                       #newCols.append(col.tostring()) 
                    #else:
                        #newCols.append(col)
                #newRows.append(newCols)
            #rows = newRows
        # end bugfix
    cursor.close ()
    conn.commit()
    return rows

def writeRow(db, table, columnData, keys=None):
    """ write row to table, columData can be a dict or a namedtuple, keys are the keys of values that are to be written from the dict/namedtuple """
    if keys==None:
        if type(columnData)==type({}):
            keys=list(columnData.keys())
        else:
            keys=columnData._fields
            columnData = columnData._asdict()
    
    values = [columnData[key] for key in keys]

    keyString = ", ".join(keys)
    placeHolders = ["%s"]*len(keys)
    placeHolderString = ", ".join(placeHolders)
    sql = "INSERT INTO %s (%s) VALUES (%s);" % (table, keyString, placeHolderString)
    cursor = db.cursor()
    cursor.execute(sql, values)
    
class SqlTableReaderGeneric:
    def _defineRowType(self):
        colNames = [desc[0] for desc in self.cursor.description]
        self.RowType = collections.namedtuple("MysqlRow", colNames)

class SqlTableReader_BigTable(SqlTableReaderGeneric):
    """ Read rows from mysql table, keep table on server and fetches row-by-row"
    >>> p = SqlTableReader_BigTable("show variables like '%protocol_version%';")
    >>> list(p.generateRows())
    [MysqlRow(Variable_name='protocol_version', Value='10')]
    """
    def __init__(self, query, conn=None, **kwargs):
        if conn==None:
            conn = sqlConnection(**kwargs)
            self.connArgs = kwargs
        self.query = query
        self.cursor = conn.cursor(MySQLdb.cursors.SSCursor)
        self.rows = self.cursor.execute(query)
        self._defineRowType()
        self.kwargs=kwargs

    def generateRows(self):
        """ fetch row from mysql server, try to re-establish connection if it times out """
        errCount=0
        while True:
            try:
                rowRaw = self.cursor.fetchone()
                if rowRaw==None:
                    break
                else:
                    rowTuple = self.RowType(*rowRaw)
                    yield rowTuple
                    
            except _mysql_exceptions.OperationalError:
                    errCount+=1
                    if errCount==100:
                        raise Exception("mysql connection error, even after retrials")
                        break
                    else:
                        logging.info("Mysql connection problems, will reconnect in 10 secs")
                        time.sleep(10)
                        self.__init__(self.query, conn=None, **self.kwargs)

class SqlTableReader(SqlTableReaderGeneric):
    """ Read rows from mysql table, will keep whole table in memory "
    >>> conn = MySQLdb.connect(read_default_file="~/.my.cnf")
    >>> p = SqlTableReader("show variables like '%protocol_version%';", conn)
    >>> list(p.generateRows())
    [MysqlRow(Variable_name='protocol_version', Value='10')]
    """
    def __init__(self, query, conn=None, **kwargs):
        if conn==None:
            conn = sqlConnection(**kwargs)
        self.cursor = conn.cursor()
        self.rows = self.cursor.execute(query)
        self.data = self.cursor.fetchall()
        self._defineRowType()

    def generateRows(self):
        for i in range(0, len(self.data)):
            rowRaw = self.data[i]
            rowTuple = self.RowType(*rowRaw)
            yield rowTuple

    def asDict(self, keyIdx=0, valueIdx=1):
        result = {}
        for row in self.generateRows():
            key = row[keyIdx]
            value = row[valueIdx]
            result[key] = value
        return result

    def writeToFile(self, fileObject):
        for row in self.generateRows():
            row = [str(d) for d in row]
            tabline = "\t".join(row)+"\n"
            fileObject.write(tabline)

def concatHeaderTabFiles(filenames, outFilename, keyColumn=None, progressDots=None):
    """ concats files and writes output to outFile, will output headers (marked with #!) only once """
    outF = openFile(outFilename, "w")
    fno = 0
    allKeys = set()
    count = 0
    for fname in filenames:
        lno = 0
        for l in open(fname):
            l = l.strip("\n")
            if keyColumn!=None:
                key = l.split("\t")[keyColumn]
                if key in allKeys:
                    logging.error("Key %s appears twice" % key)
                    exit(1)
            if lno!=0 or (lno==0 and fno==0 and l.startswith("#")):
                outF.write(l)
                outF.write("\n")
            lno+=1
            count+=1
            if progressDots!=None and count% progressDots==0:
                print(".", end=' ')
                sys.stdout.flush()
        fno+=1

splitReOp = re.compile("[MIDNSHP]")
splitReNumbers = re.compile("[0-9]+")
def samToBed(row):
    strand = (row.flag & 16) == 16
    if strand==1:
        strand="-"
    else:
        strand="+"

    if row.pos==0:
        return None

    featStart = row.pos -1

    cigarLengths = [int(x) for x in splitReOp.split(row.cigar) if x!=""]
    cigarOps     = [x for x in splitReNumbers.split(row.cigar) if x!=""]
    cigarLine    = list(zip(cigarOps, cigarLengths))
    assert(len(cigarLengths)==len(cigarOps)==len(cigarLine))

    featLength = 0
    matches = 0
    for op, length in cigarLine:
        length= int(length)
        if op in "MDNP":
            featLength+=length
        if op=="M":
            matches += length
    featEnd = featStart+featLength

    return row.rname, featStart, featEnd, row.qname, matches, strand
    
def parseSam(fileObj):
    tp = TableParser(fileObj=fileObj, fileType="sam")
    for row in tp.generateRows():
        yield samToBed(row)

def parseDict(fname, comments=False, valField=1, doNotCheckLen=False, otherFields=False, headers=False, keyType=None, valueType=None, errorSummary=False, inverse=False):
    """ parse file with key -> value pair on each line, key/value has 1:1 relationship"""
    """ last field: set valField==-1, return as a dictionary key -> value """
    if fname==None or fname=="":
        return {}
    dict = {}
    f = openFile(fname)
    errors = 0
    if not f:
        return dict

    if headers:
        f.readline()
    for l in f:
        #print l
        fs = l.strip().split("\t")
        if comments and l.startswith("#"):
            continue
        if not len(fs)>1:
            if not doNotCheckLen:
                sys.stderr.write("info: not enough fields, ignoring line %s\n" % l)
                continue
            else:
                key = fs[0]
                val = None
        else:
            key = fs[0]

            if keyType:
                try:
                    key = keyType(key)
                except:
                    logging.warn("Line %s: could not cast key to correct type" % l)

            if not otherFields:
                val = fs[valField]
            else:
                val = fs[1:]
            
            if valueType:
                try:
                    val = valueType(val)
                except:
                    logging.warn("Line %s: could not cast value to correct type" % l)

        if inverse:
            key, val = val, key

        if key not in dict:
            dict[key] = val
        else:
            if errorSummary:
                errors+=1
            else:
                sys.stderr.write("info: file %s, hit key %s two times: %s -> %s\n" %(fname, key, key, val))
    if errorSummary:
        logging.warn("found %d lines with non-unique key-value associations!" % errors)
    return dict

def openBed(fname, fileType="bed3"):
    "return iterator for bed files "
    fh = openFile(fname)
    return TableParser(fh, fileType=fileType).lines()

def makeTableCreateStatement(tableName, fields, type="sqlite", intFields=[], primKey=None, idxFields=[], inlineIndex=False, primKeyIsAuto=False, fieldTypes={}):
    """
    return a tuple with a create table statement and a list of create index statements.
    returns the index statements separately for additional speed.
    if inlineIndex is true, index create is part of the table create statement.

    >>> makeTableCreateStatement("testTbl", ["test", "hi", "col3"], intFields=["hi"], primKey="test", idxFields=["col3"])
    ('CREATE TABLE IF NOT EXISTS testTbl (test TEXT PRIMARY KEY, hi INTEGER, col3 TEXT); ', 'CREATE INDEX testTbl_col3_idx ON testTbl (col3);')
    """
    intFields = set(intFields)
    idxFields = set(idxFields)
    idxFieldNames = [] 
    parts = []
    idxSqls = []
    for field in fields:
        if type=="sqlite":
            ftype = "TEXT"
        else:
            ftype = "VARCHAR(255)"

        if field in fieldTypes:
            ftype = fieldTypes[field]
        elif field in intFields:
            ftype = "INTEGER"

        if field in idxFields:
            idxSqls.append("CREATE INDEX IF NOT EXISTS %s_%s_idx ON %s (%s);" % \
                (tableName, field, tableName, field))
            idxFieldNames.append(field)

        statement = field+" "+ftype
        if field == primKey:
            statement += " PRIMARY KEY"
            if primKeyIsAuto:
                statement += " AUTO_INCREMENT"
                
        parts.append(statement)

    idxStr = ""
    if inlineIndex:
        for idxFieldName in idxFieldNames:
            idxStr += ", INDEX (%s)" % idxFieldName

    tableSql = "CREATE TABLE IF NOT EXISTS %s (%s %s); " % (tableName, ", ".join(parts), idxStr)
    return tableSql, idxSqls

def loadTsvSqlite(dbFname, tableName, tsvFnames, headers=None, intFields=[], \
    primKey=None, idxFields=[], dropTable=True):
    " load tabsep file into sqlLite db table "
    # if first parameter is string, make it to a list
    if len(tsvFnames)==0:
        logging.debug("No filenames to load")
        return
    if isinstance(tsvFnames, str):
        tsvFnames = [tsvFnames]

    if os.path.isfile(dbFname):
        lockDb = False
        finalDbFname = None
    else:
        lockDb = True
        finalDbFname = dbFname
    con, cur = openSqlite(dbFname, lockDb=lockDb)

    # drop old table 
    if dropTable:
        logging.debug("dropping old sqlite table")
        cur.execute('DROP TABLE IF EXISTS %s;'% tableName)
        con.commit()

    # create table
    createSql, idxSqls = makeTableCreateStatement(tableName, headers, \
        intFields=intFields, idxFields=idxFields, primKey=primKey)
    logging.log(5, "creating table with %s" % createSql)
    cur.execute(createSql)
    con.commit()

    logging.info("Loading data into table")
    tp = maxCommon.ProgressMeter(len(tsvFnames))
    sql = "INSERT INTO %s (%s) VALUES (%s)" % (tableName, ", ".join(headers), ", ".join(["?"]*len(headers)))
    for tsvName in tsvFnames:
        logging.debug("Importing %s" % tsvName)
        if os.path.getsize(tsvName)==0:
            logging.debug("Skipping %s, zero size" % tsvName)
            continue
        rows = list(maxCommon.iterTsvRows(tsvName))
        logging.log(5, "Running Sql %s against %d rows" % (sql, len(rows)))
        cur.executemany(sql, rows)
        con.commit()
        tp.taskCompleted()

    logging.info("Adding indexes to table")
    for idxSql in idxSqls:
        cur.execute(idxSql)
        con.commit()

    con.close()

    if finalDbFname!=None:
        logging.info("moving over ramdisk db to %s" % dbFname)
        shutil.move(dbFname, finalDbFname)

def insertSqliteRow(cur, con, tableName, headers, row):
    " append a row to an sqlite cursor "
    sql = "INSERT INTO %s (%s) VALUES (%s)" % (tableName, ", ".join(headers), ", ".join(["?"]*len(headers)))
    logging.log(5, "SQL: %s" % sql)
    cur.execute(sql, row)
    con.commit()

def openSqliteCreateTable(db, tableName, fields, intFields=None, idxFields=None, primKey=None, retries=1):
    " create the sqlite db and create a table if necessary "
    con, cur = openSqlite(db, retries=retries)
    if primKey==None:
        primKey = fields[0]
    createSql, idxSqls = makeTableCreateStatement(tableName, fields, \
        intFields=intFields, idxFields=idxFields, primKey=primKey)
    logging.log(5, "creating table with %s" % createSql)
    cur.execute(createSql)
    con.commit()
    for idxSql in idxSqls:
        cur.execute(idxSql)
    con.commit()
    return con, cur

def namedtuple_factory(cursor, row):
    """
    Usage:
    con.row_factory = namedtuple_factory
    """
    fields = [col[0] for col in cursor.description]
    Row = collections.namedtuple("Row", fields)
    return Row(*row)

def openSqlite(dbName, asNamedTuples=False, lockDb=False, timeOut=10, retries=1, asDict=False):
    " opens sqlite con and cursor for quick reading "
    logging.debug("Opening sqlite db %s" % dbName)
    tryCount = retries
    con = None
    while tryCount>0 and con==None:
        try:
            con = sqlite3.connect(dbName, timeout=timeOut)
        except sqlite3.OperationalError:
            logging.info("Database is locked, waiting for 60 secs")
            time.sleep(60)
            tryCount -= 1

    if asNamedTuples:
        con.row_factory = namedtuple_factory
    elif asDict:
        con.row_factory = sqlite3.Row

    cur = con.cursor()
    #cur.execute("PRAGMA read_uncommited=true;") # has only effect in shared-cache mode
    con.commit()
    if lockDb:
        cur.execute("PRAGMA synchronous=OFF") # recommended by
        cur.execute("PRAGMA count_changes=OFF") # http://blog.quibb.org/2010/08/fast-bulk-inserts-into-sqlite/
        cur.execute("PRAGMA cache_size=800000") # http://web.utk.edu/~jplyon/sqlite/SQLite_optimization_FAQ.html
        cur.execute("PRAGMA journal_mode=OFF") # http://www.sqlite.org/pragma.html#pragma_journal_mode
        cur.execute("PRAGMA temp_store=memory") 
        con.commit()
    return con, cur

def iterSqliteRows(db, tableName):
    """
    yield rows from sqlite db
    """
    con, cur = openSqlite(db, asNamedTuples=True)
    for row in cur.execute("SELECT * FROM %s" % tableName):
        yield row

def iterSqliteRowNames(cur, tableName):
    for row in cur.execute("PRAGMA table_info(%s);" % tableName):
        yield row[1]

if __name__ == "__main__":
    import doctest
    doctest.testmod()




from os.path import expanduser, join, isdir, isfile, normpath, dirname, abspath
from os import makedirs, getenv
import logging
from .maxCommon import getAppDir

# first parse the user config file. The variables will later be re-applied
# after the setting in this file
confName = getenv("PUBMUNCH_CONF")
if (confName is not None) and not isfile(confName):
    raise Exception("configuration file specified in PUBMUNCH_CONF environment variable ({}) does not exist".format(confName))
if confName is None:
    confName = expanduser("~/.pubConf")
newVars = {}
if isfile(confName):
    exec(compile(open(confName).read(), confName, 'exec'), {}, newVars)
    for key, value in newVars.items():
        locals()[key] = value

# GENERAL SETTINGS   ================================================
# baseDir for internal data, accessible from cluster 
# used for data created during pipeline runs
# Preference is given to a locally defined directory
if "pubsDataDir" not in locals():
    pubsDataDir = "."

# static data, accessible from cluster, but part of distribution
# these are basic files like gene lists, marker lists, journal lists
# Some of it can be updated with pubPrepXXX commands
staticDataDir = join(getAppDir(), "data")

# scripts only used at UCSC
ucscScriptDir = normpath(join(dirname(__file__), "..", "ucscScripts"))

# external tools
extToolDir = normpath(join(dirname(__file__), "..", "ext"))

# a directory with files that associate publishers with journals
# one of them is the NLM Catalog, others we got from publishers or created them semi-
# manually
#journalListDir = join(staticDataDir, "journalLists")
#journalListDir = join(staticDataDir, "journalLists")
journalListDir = "/hive/data/inside/pubs/journalLists/"

# base dir for anything journal related in the repo data dir
journalInfoDir = join(staticDataDir, "journals")

# the lists are reformatted into this table. It is created by pubJournals and used by pubPrepCrawl
# it contains the ISSNs and server names for each publisher
# it is required by pubCrawl for the highwire configuration
publisherIssnTable = join(journalInfoDir, "publishers.tab")

# same info, but one line per journal
# not used anymore
journalTable = join(journalInfoDir, "journals.tab")


# directory with various tracking files expected vs retrieved documents
inventoryDir = join(pubsDataDir, "inventory")

# a directory on the main server (hgwdev) on which the scripts is run, on local disk
# with lots of space available. Used to store intermediate results ? -> pubClassify
# localHeadDir = '/scratch/max/pubTools'

# DB PARSER SETTINGS ================================================
# directory for files with parsed DBs each DB, e.g. uniprot or pdb
dbRefDir = '/hive/data/inside/pubs/parsedDbs'

# directories with a local copy of PDB and uniprot, ncbi genes, refseq
uniProtBaseDir = '/hive/data/outside/uniProtCurrent'
pdbBaseDir = '/hive/data/outside/pdb'
ncbiGenesDir = '/hive/data/outside/ncbi/genes/'
#ncbiRefseqDir = '/hive/data/outside/ncbi/refseq/release/vertebrate_mammalian'
ncbiRefseqDir = '/hive/data/outside/ncbi/refseq/H_sapiens/mRNA_Prot'

# VARIANT MAPPING ========================================================
# directory for files that are needed to build the variants at UCSC but not needed 
# to run the programs. Examples are the PSL alignments built at UCSC, before they are
# indexed into sqlite
varBuildDir = '/hive/data/inside/pubs/variants'

# OMIM genemap2 on local disk or via internet
omimUrl = "file:///hive/data/outside/omim/01122013/genemap2.txt"

# CONVERTER SETTINGS ================================================
# for auto mode: base data dir with incoming input files from publishers or databases
extDir = '/hive/data/outside/pubs/'

# for auto mode: base dir of all publisher text files
textDir = join(pubsDataDir, "text")

# gzip compress the text data? Seeking is faster if not gzip compressed.
compress = True

# for pubConvMedline:

# an sqlite db with the content of medline, kept up-to-date
medlineDbPath = join(textDir, "medline", "medline.db")

# CRAWLER SETTINGS ==================================================

# only needed for Elsevier crawling.
# get your own key from 
# https://www.elsevier.com/about/company-information/policies/text-and-data-mining
# then paste it here
elsevierApiKey = None

# the useragent to use for http requests
httpUserAgent = 'genomeBot/0.1 (YOUREMAIL, YOURWEB, YOURPHONE)'

# how long to wait for DNS queries, TCP connects, between retries and TCP reads, in seconds
httpTimeout = 5

# how to long wait for the downloading of files, in seconds
# after this the download will be canceled and the paper marked as unsuccessful
httpTransferTimeout = 60

# if you need to use a proxy to access journals, set it here
httpProxy = None

# SFX server for pubmed entries without an outlink
# YOU NEED TO DEFINE THIS IF YOU WANT TO USE SFX
# the UC SFX server is "http://ucelinks.cdlib.org:8888"
# the UR SFX server is "http://sfx.bib-bvb.de/sfx_ubr"
# but please use your own SFX server at your university
# google for it or ask your library for the URL
crawlSfxServer = None

# you can set special delay times here. E.g. to make all npg journals
# get crawled slowly at 30secs, set "npg" to 30
crawlDelays = {}

# in some cases, the automatic publisher<->ISSN assignment has errors
# this is either because the journal has switched publishers or
# because it is simply missing in the journal lists that we have.
# For these cases you can add or re-assign a given ISSN to a directory.
# format: ISSN -> (directory, yearStart, yearEnd)
# yearStart and yearEnd can be None if you want all years
# Any ISSN defined here is not automatically assigned to publishers anymore
# only the manual assignments are used for these ISSNs

crawlIssnOverwrite = {
# The journal of hypertension has been transferred to NPG after 2008
"1941-7225" : ("sciencedirect", 1995, 2008),
"1941-7225" : ("npg", 2008, 2099)
}

# a list of file extensions to download when crawling for supplementary files
# better not .mov,.vid or .avi, nor html (=might get normal webpages)
# you can add additional types on a per-publisher basis with "addSuppFileTypes" in the pubCrawler host config
crawlSuppExts = set(['gif', 'svg', 'tiff', 'tif', 'jpg', 'xls', 'doc', 'pdf', 'ppt', 'asc', 'txt', 'csv', 'tsv', 'tab', 'eps'])

# DOWNLOAD SETTINGS ==================================================

# for Elsevier updates:
# We need the URL to the Consyn Update RSS feed. 
# Login to consyn.elsevier.com, click on batches/RSS feed, paste the URL here
# they look like https://consyn.elsevier.com/batch/atom?key=XXXXXXXXXXXXXXXXXXX
# (at UCSC: this is defined in ~/.pubTools.conf)
consynRssUrl = "YOURURL"

# for springer updates:
# we got a username / password from DDS Support in Heidelberg
# (at UCSC: these are defined in ~/.pubTools.conf)
springerUser = ""
springerPass = ""

# GENERAL PUBLICATION FILE CONFIG SETTINGS ============================

# which dataset should be loaded by the "load" step in pubMap ?
# loadDatasets = ["elsevier", "pmc", "crawler", "springer"]

# bundles are set of datasets and annotations run regularily over them
bundleToText = {
    "pubsTrack" : ["pmc", "springer", "elsevier", "crawler"],
}


TEMPDIR = "/tmp/pubTools" # local filesystem on cluster nodes
FASTTEMPDIR = TEMPDIR

maxBinFileSize = 50000000 # maximum filesize of any file before conversion to ASCII
# pretty big, to allow big blobs like TIFF files etc, in docx files
maxTxtFileSize = 10000000 # maximum filesize of any file after conversion to ASCII
minTxtFileSize = 60 # minimum filesize of any file after conversion to ASCII
# 60 to allow files with just a figure legend to be processed
mapReduceTmpDir = pubsDataDir + "/mapReduceTemp" # cluster-wide directory to collect results

# parasol batches dir
clusterBatchDir = pubsDataDir + "/runs/"
# the base directory for text repository directories
textBaseDir = pubsDataDir + "/text/"
# the central base directory for all text annotations
annotDir = pubsDataDir + "/annot/"
# central directory for exported fasta file
faDir = pubsDataDir + "/fastaExport/"

# all logfiles
logDir = pubsDataDir + "/log/"

# head node of cluster
clusterHeadNode = "ku.sdsc.edu"
# type of cluster, either parasol or sge or local (single CPU) or localhost:5 for five CPUs
clusterType = "localhost"

# _sourceDir = "/cluster/home/max/projects/pubs/tools"
_sourceDir = getAppDir()

# base directory for searcher algorithm code, like regex annotation, dna annotation, etc
scriptDir = join(_sourceDir,"taggers")

# cmdLine to start jython, used to run java annotators
jythonCmd= "/cluster/home/max/software/jre1.7.0/bin/java -jar "+dirname(__file__)+"/jython.jar"

# assignment of pubMap pipeline steps to cluster machines
# pubMap will ssh into a machine to run these steps
stepHosts = {"sortCdna" : "localhost", "sortProt" : "localhost", "sortGenome" : "localhost"}

# email for ncbi eutil requests and error email by pubCrawl
# at UCSC: overriden by local .pubConf
email = ""

# how much to wait between two eutils requests
eutilsDelaySecs = 3

# identifiers for articles, files and annotations are all in 64bit space.
# a certain number of digits are used for articles, files and annotations
# 10 digits for articles,
# 3 for files 
# 5 for annotations
# that means that each publisher cannot have more than one billion articles,
# one article not more than 1000 files and one algorithm cannot return more than
# 100.000 annotations per file
ARTICLEDIGITS=10 # number of digits to use for annotation ID 
FILEDIGITS=3 # number of digits to use for annotation ID 
ANNOTDIGITS=5 # number of digits to use for annotation ID 

# articleIds should start at which position in our namespace?
# this is to make sure that articleIds are unique
identifierStart = {
    "pmc"      : 1000000000,
    "elsevier" : 2000000000,
    "springer" : 2500000000,
    "medline"  : 3000000000,
    "bing"     : 3500000000,
    "genbank"  : 4000000000,
    "imgt"     : 4300000000,
    "pdfDir"   : 4400000000,
    "yif"      : 4500000000,
    "crawler"  : 5000000000,
    "free"     : 6000000000  # to indicate end of list, always keep this here
    # the "free" entry is necessary to indicate the range of the 2nd to last entry
}

extToolDir = _sourceDir+"/external"

# commands to convert various filetypes to ascii text 
# $in and $out will be replaced with temp filenames
# pdf2 is only used if output of pdf contains more than 10 unprintable characters
# as pdfbox is quite slow
# you can define variables and use them, see extToolDir
CONVERTERS = {
    "doc":"catdoc $in > $out",
    "docx":"%(extToolDir)s/docx2txt-1.2/docx2txt.pl < $in > $out",
    "xls":"xls2csv $in > $out",
    "xlsx":"ssconvert $in $out",
    "ppt":"catppt $in > $out",
    "htm":"html2text $in  --unicode-snob --images-to-alt --ignore-links --ignore-emphasis > $out",
    "html":"html2text $in  --unicode-snob --images-to-alt --ignore-links --ignore-emphasis > $out",
    #"htm":"html2text -style pretty -nobs $in | tr -s ' ' > $out",
    # 'pretty' avoids **** markup for <h3> section names
    #"html":"html2text -style pretty -nobs $in | tr -s ' ' > $out",
    #"htm":"links -dump $in -dump-charset utf8 > $out",
    #"html":"links -dump $in -dump-charset utf8 > $out",
    "csv":"COPY",
    "txt":"COPY",
    "asc":"COPY",
    "xml":"XMLTEXT",
    "nxml":"NXMLTEXT",
    "pdf":"pdftotext -q -nopgbrk -enc UTF-8 -eol unix $in $out",
    "pdf2":"java -Xmx512m -jar %(extToolDir)s/pdfbox-app-1.6.0.jar ExtractText $in $out -encoding utf8"
}

# sometimes (e.g. if downloaded from the web) we don't have a file extension, but only 
# a mime-type. The following list maps from mime types to file extensions = converters
MIMEMAP = {
    "application/msword":"doc",
    "application/vnd.ms-excel":"xls",
    "application/vnd.ms-powerpoint":"ppt",
    "text/html":"html",
    "text/csv":"csv",
    "text/tab-separated-values":"csv",
    "text/plain":"txt",
    "application/xml":"xml",
    "application/pdf":"pdf"
}

# override the weird highwire delay times for certain URLs with lower values
# you need to check with Highwire on these first
# at ucsc, this is overriden with our local .pubConf 
highwireDelayOverride = {
}

# when splitting text files for cluster jobs,
# how many "chunks" should be created?
# chunkCount = 2000 # XXX  not used 
# when creating chunks, how many files should go into one chunk?
# used by pubConvCrawl
chunkArticleCount = 400

# for conversion jobs: how many cluster jobs should we run in parallel ?
# can be used to limit I/O and be nice to other cluster users
convertMaxJob = 200

# PMID RESOLVER

# filename of pubCompare output file in source dataset directory
idFname = "medline.ids.tab"

# GENBANK CONVERSION

# genbank divisions to run on
genbankDivisions = ["inv","mam","pat","pln","pri","rod","una","vrt"]

# extract sequences for these species from Genbank
genbankTaxons = ["Homo sapiens", "Mus musculus", "Drosophila melanogaster"]

# ignore sequences that are longer than x bp (e.g. BACS)
genbankMaxLen = 40000

# maximum number of accession for a publication reference
# all submission with more sequences than this are skipped
genbankMaxRefCount = 50

# DNA MAPPING / GENOME BLATTING SETTINGS ============================

pubMapBaseDir = "./map/"

# directory for exported cdr3 files
cdr3Dir = pubMapBaseDir + "cdr3Export/"

# this is the genbank mapping config file by Mark Diekhans' pipeline
# it is required for genome partitioning 
# current one can be downloaded from:
# http://genome-source.cse.ucsc.edu/gitweb/?p=kent.git;a=blob_plain;f=src/hg/makeDb/genbank/etc/genbank.conf;hb=HEAD
GBCONFFILE     = "/cluster/data/genbank/etc/genbank.conf"

# these variables assign genomes to keywords in text files
# the orgDetect.py plugin will create annotations on the text files
# for all of these keywords
# pubMap will then blat sequences from a textfile only on genomes
# for which a keyword has been found
# ALL SPECIES that the pipeline needs to blat on need to be defined here
# even if no species recognition for them is performed

speciesNames = {
'hg19' : ['human', 'sapiens', ' homo ', ' Homo ', 'patient', 'cell line','cell culture'],
'mm10' : ['mouse', 'musculus', 'rodent'],
'rn4' : [' rat ', 'novegicus', 'rodent'],
#'nonUcsc_archaea' : [],
'danRer7' : ['zebrafish', 'rerio', 'Danio'],
'xenTro2' : [' xenopus', 'Xenopus', 'tropicalis', 'laevis'],
'susScr3' : [' swine ', ' swines ', ' pigs ', ' pig ', ' porcine ', ' scrofa '],
'bosTau7' : [' cattle ', ' cows ', ' cow ', ' bovine ', ' beef ', ' bovis '],
'galGal4' : [' chicken ', ' poultry ', ' chickens ', ' gallus '],
'oryLat2' : ['medaka', 'Medaka', 'latipes', 'Oryzias'],
'dm3' : [' fruitfly ', ' flies ', 'melanogaster', 'Drosophilids', ' fruitflies '],
'ce10' : ['elegans', 'Caenorhabditis', 'nematode', ' worms'],
'ci2' : ['ascidian', 'intestinalis', 'chordates', 'Ciona'],
'sacCer2' : ['cerevisiae', 'Saccharomyces', 'yeast'],
'felCat5' : ["domestic cat", "felis cattus", "felidae", " cats "],
'nonUcsc_arabidopsisTair10' : ['arabidopsis', 'Arabidopsis', 'thaliana', 'thale cress'],
'nonUcsc_Pfalciparum3D7' : ['plasmodium', 'Plasmodium', 'falciparium', 'malaria'],
'nonUcsc_grapevine12x' : [' grapevine ', ' vitis ', 'pinot noir', ' vigne'],
'nonUcsc_bacteria' : ['bacteria ', 'bacterial ', 'microbiology', "coli ", "prokaryotes"],
'nonUcsc_viral' : [' virus ', 'viral '],
'nonUcsc_fungi' : ['Aspergillus'],
'nonUcsc_dnasu' : ['plasmid', 'cloned into', 'cloning vector', 'restriction site'],
'nonUcsc_biobricks' : ['synthetic biology', 'biobricks'],
'nonUcsc_wolb1' : ['wolbachia']
#'nonUcsc_dnasu' : [' plasmid ']
}

# path for genome files that start with 'nonUcsc_'
# this directory has to contain a geneBank.conf file
# in UCSC format to define parameters for these
# assemblies
nonUcscGenomesDir = pubsDataDir+"/nonUcscGenomes"

# During best-genome filtering, sometimes two genomes score equally
# if this is the case, pick the best one, in this order.
# (the first one has highest priority)

# if a genome is not part of the genbank config system then you need to prefix it with
# "nonUcsc_". Any db name like this will be searched in nonUcscGenomesDir (see below).
# e.g. nonUcsc_archaea will be resolved to /hive/data/inside/pubs/nonUcscGenomes/archaea.2bit
# The BLAT-wrapper also needs a .ooc file with the same name.
#alignGenomeOrder = ['hg19', 'mm10', 'rn4', 'nonUcsc_archaea', 'danRer7', 'dm3',
alignGenomeOrder = ['hg19', 'mm10', 'rn4', 'danRer7', 'nonUcsc_wolb1', 'dm3',
'xenTro2', 'oryLat2', 'susScr3', 'bosTau7', 'galGal4', 'felCat5', 'ci2', 'ce10', 'sacCer2',
'nonUcsc_arabidopsisTair10', 'nonUcsc_Pfalciparum3D7', 'nonUcsc_grapevine12x',
'nonUcsc_bacteria', 'nonUcsc_viral', 'nonUcsc_fungi', 'nonUcsc_dnasu', "nonUcsc_biobricks"
]

# these genomes are used if no species name matches are found
#defaultGenomes = ["hg19", "mm9", "rn4", "danRer7", "dm3", "ce10", "nonUcsc_archaea"]
defaultGenomes = ["hg19", "mm10", "rn4", "danRer7", "dm3", "ce10", "nonUcsc_bacteria"]

# these genomes are always added, no matter which species names are found
# human might not be recognized as a name for cell lines
# bacteria have so many names that we don't recognize them
# same for plasmids
#alwaysUseGenomes = ["hg19", "nonUcsc_archaea", "nonUcsc_dnasu"]
#alwaysUseGenomes = ["hg19", "nonUcsc_archaea"]
alwaysUseGenomes = ["hg19"]


# for some genomes we don't have refseq data
#noCdnaDbs = ["sacCer2", "nonUcsc_archaea", "nonUcsc_arabidopsisTair10", "nonUcsc_Pfalciparum3D7"]
noCdnaDbs = ["sacCer2", "nonUcsc_arabidopsisTair10", "nonUcsc_Pfalciparum3D7"]

# some text datasets are just variants of others
# for these, to avoid annotation id overlaps with the main dataset
# we add some offset to their annotation IDs
specDatasetAnnotIdOffset = {"yif" : 12000 }

# each species in alignGenomeOrder has to be in speciesNames
if not len(set(alignGenomeOrder).intersection(speciesNames))==len(speciesNames):
    logging.error("missing in speciesNames: %s" % repr(set(alignGenomeOrder)-set(speciesNames)))
    logging.error("missing in alignGenomeOrder: %s" % repr(set(speciesNames)-set(alignGenomeOrder)))
    raise Exception("illegal species name configuration")

assert(len(set(defaultGenomes).intersection(speciesNames))==len(defaultGenomes))
assert(len(set(alwaysUseGenomes).intersection(defaultGenomes))==len(alwaysUseGenomes))

# minimum size of dna sequence to be considered for blatting
minSeqLen=17
# minimum size of protein sequence to be considered for blatting
minProtSeqLen=7
# maximum size of dna or protein sequence
maxSeqLen=50000

# maximum size of sequence to be considered "short", sequences can be "short" or "long"
# 50 because: two neighboring primers in a table are recognized as one long primer.
# Blat will split it, but it has to fall into the short category as otherwise
# each individual match won't exceed the 25bp threshold.
shortSeqCutoff = 50

# the sequences from papers will be split into several fa files per organism
# maximum size of fasta file per organism (to keep hippos low, short/long separated)
queryFaSplitSize = {"long" : 8000000, "short" : 400000 }

# how to split the genomes? used for bigBlat.py
genomeSplitParams = { "winSize" : 3000000, "overlap":100000 }

#  same, but for Cdna and protein blatting
cdnaFaSplitSizes = {"long" : 12000000, "short" : 8000000 }
protFaSplitSizes = {"long" : 12000000}

# biggest chunk of split cDna target files
cdnaSplitSize = 6000000
protSplitSize = 6000000

# options for blatting and pslFiltering for each sequence type
# original values were:
# minAli=0.7,nearTop=0.01,minNearTopSize=18,ignoreSize,noIntrons
# short minAli=0.7,nearTop=0.01
seqTypeOptions = {
    "short": ("repeats=lower,stepSize=5,minScore=16,minMatch=1,maxIntron=5", "minQSize=17,minNonRepSize=15,ignoreNs,bestOverlap,minId=0.95,minCover=0.15,globalNearBest=0.01"),
    "long":  ("repeats=lower,minScore=25", "minQSize=18,minNonRepSize=16,ignoreNs,bestOverlap,minId=0.95,minCover=0.15,globalNearBest=0.01")
}

# proteins have only "long" sequences = 6 amino acids
protBlatOptions = {
    "long":  ("t=dnax,q=prot", "minQSize=6,minNonRepSize=6,ignoreNs,bestOverlap,minId=0.95,minCover=0.15,globalNearBest=0.01")
}
# distance for chaining psls from same article into bed
maxChainDist={'default': 50000, 'sacCer2' : 5000}

# maximum number of matches for an article per DB, articles with 
# more matches will be ignored (before chaining)
maxDbMatchCount=1000

# ignore articles with more than a certain number of genomic featuress (after chaining)
maxFeatures = 50

# how many bp does a chain has to cover?
# smaller chains will be considered "noise" and skipped
minChainCoverage=23

# maximum length of chain
# this can sometimes happen if a chain contains many individual
# matches spread evenly over a very long locus
# we eliminate these very long chains, they are likely to be 
# not very informative
maxChainLength=3000000

# when splitting the big psl file, how many chunks should be fused into one?
# this is used to separate the big psl file onto smaller cluster jobs
# e.g. when this is 10 and we have 2000 total chunks, the chaining 
# will run on 200 pieces 
chunkDivider = 10

# which alignment table shall we use for cdna mapping?
# (this is used by pubPrepCdnaDir to download mRNA alignmenst 
# and mRNA fasta files
cdnaTable = 'refSeqAli'

# pubMap static data files directory, used below
_pubsDataDir = "/hive/data/inside/pubs/pubMapData"
# where to store the cDNA data (refseq alignments and fasta files)
cdnaDir = _pubsDataDir + "/cdnaDb"

# where to store the loci data (regions around longest transcripts for genes,
# their symbols and entrez IDs) a directory with <db>.bed files that associate
# genomic locations with the closest entrez gene and symbol loci creation also
# requires ncbiGenesDir
lociDir = _pubsDataDir+"/loci"

# file with impact factors for bed annotation
impactFname = join(staticDataDir, "map", "impact2011.tab")

# directory with the files t2g.sql, t2gArticle.sql and t2gSequence.sql
# required for loading into data UCSC genome browser
# (for "pubMap <dataset> load")
sqlDir = "/cluster/home/max/projects/pubs/tools/sql/"

# when writing tables for mysql, we cut most columns to a maximum size,
# to account for data errors (elsevier)
maxColLen = 255


# MARKER MAPPING ============================
markerDbDir = "/hive/data/inside/pubs/markerDb/"
humanDb = "hg19"

hgncFname = "/hive/data/outside/hgnc/111413/hgnc_complete_set.txt"

# directory with subdirs, one per db, that contains chrom.sizes files
genomeDataDir = "/hive/data/genomes"

# BLASTP
#blastBinDir = "/cluster/bin/blast/x86_64/blast-2.2.20/bin"
blastBinDir = '/hive/data/outside/blast229'

# default Mysql database 
mysqlDb = 'publications'

# UNIPROT PARSING ============================
# only convert records with one of these taxon Ids
uniProtTaxonIds = [9606, 10090, 10116, 7955]


# CLASSIFICATION ==============================
# directory with the SVMlight binaries
svmlBinDir = "/hive/data/inside/pubs/svmlight"

# directory to write html output files, one per DB
classOutHtmlDir = "/cluster/home/max/public_html/mining/classes"
testOutHtmlDir = "/cluster/home/max/public_html/mining/testClasses"

# descriptions for the various classes based on lists of PMIDs from DBs
classDescriptions = {
'gwas' : "Genome-wide association studies",
'redfly' : "Drosophila cis-regulatory assays",
'aptamerBase' : "Aptamers",
'chembl' : "Drug-like small molecules",
'chimerDb' : "Fusion genes",
'clinicalTrials' : "Clinical Trials",
'cosmic' : "Cancer genes and mutations",
'flybase' : "Drosophila Genetics",
'mgi' : "Mouse Genetics",
'omim' : "Human Genetics",
'pdb' : "Protein Structures",
'pharmGKB' : "Genetic variation and drug response",
'reactome' : "Protein interactions",
'wormbase' : "C. elegans genetics",
}

# the final output of the classification step, for pubMap integration
classFname = join(pubsDataDir, "classify", "crawler-elsevier-pmc", "docClasses.tab")

# GENE AND MUTATION RECOGNIZERS ===========================

# directory with lists of accessions
accDataDir = join(staticDataDir, "accessions")

# directory with lots of data about genes
geneDataDir = join(staticDataDir, "genes")

# directory with data required to detect variants
varDataDir = join(staticDataDir, "variants")

# the british national corpus is a list of 30k common words in English
# used for symbol filtering
bncFname = '/hive/data/outside/pubs/wordFrequency/bnc/bnc.txt'

# now overwrite all variables with those defined in local 
# config file ( see start of this file )
for key, value in newVars.items():
    locals()[key] = value

# SOLR =======
solrUrl="http://hgwdev.soe.ucsc.edu:8983/solr"

# ACCESS METHODS (convenience) ============================

debug = False

import sys, logging, time, random

def getConverters():
    return CONVERTERS

def getFastTempDir():
    " some fast local temp, possibly a ramdisk "
    if not isdir(FASTTEMPDIR):
        makedirs(FASTTEMPDIR)
    return FASTTEMPDIR

def getTempDir():
    " create temp dir, try to resolve race conditions "
    if not isdir(TEMPDIR):
        time.sleep(random.randint(1,5)) # make sure that we are not all trying to create it at the same time
        if not isdir(TEMPDIR):
            try:
                makedirs(TEMPDIR)
            except OSError:
                logging.info("Ignoring OSError, directory %s seems to exist already" % TEMPDIR)
    return TEMPDIR

def getUcscScriptDir():
    dirname(__file__)
    
def getMaxBinFileSize():
    return maxBinFileSize

def getMaxTxtFileSize():
    return maxTxtFileSize

def mayResolveTextDir(dataDir):
    """ check if dataDir is a subdirectory of textBaseDir. If not, check if it's a valid path. "
    Return the absolute path or None if neither case is true.
    """
    inName = dataDir
    dataDir2 = join(textBaseDir, dataDir)
    if isdir(dataDir2) and not abspath(dataDir2)==abspath(textBaseDir):
        logging.debug("Resolved dataset name %s to global dataset directory %s" % (inName, dataDir))
        return abspath(dataDir2)

    if isdir(dataDir) and not abspath(dataDir)==abspath(textBaseDir):
        logging.debug("Resolved dataset name %s to directory %s" % (inName, dataDir))
        return abspath(dataDir)

    return None

def resolveTextDir(dataDir, mustFind=True):
    " check if dataDir exists, if not: try subdir with this name of textDir. abort if not found "
    fullPath = mayResolveTextDir(dataDir)
    if mustFind and fullPath == None:
        raise Exception("Could not resolve dataset %s to a directory" % dataDir)
    logging.debug("Resolved dataset name %s to dataset directory %s" % (dataDir, fullPath))
    return fullPath

def resolveTextDirs(dataString):
    " like resolveTextDir but accepts comma-sep strings and yields many "
    dirs = []
    for dataSpec in dataString.split(","):
        dirs.append( resolveTextDir(dataSpec) )
    return dirs

def getStaticDataDir():
    """ returns the data dir that is part of the code repo with all static data, e.g. train pmids
    """
    return staticDataDir

def defaultInOutDirs(datasetName):
    " return the default input and the default output directory for a dataset "
    return join(extDir, datasetName), join(textBaseDir, datasetName)

#def getDataDir(name):
    #" somewhat more flexible, can be redirected based on program name "
    #if name=="genes":
	#return join(staticDataDir, "mutFinder")
    #else:
        #assert(False)




# Classes to read and write documents to directories

# gzipfiles are stored in "chunks" of x documents, usually to get around 2000
# chunks per dataset

# reader class gets path/xxxx and will open path/xxxx.articles.gz and path/xxxx.files.gz
# reader class yields files from these two files

# writer class writes to local filesystem, as two different textfiles:
# xxxx.files, xxxx.articles
# then gzips them and copies gzipfiles to shared cluster filesystem

import os, logging, sys, collections, time, codecs, shutil, tarfile, csv, glob, operator, types
import zipfile, gzip, re, random, tempfile, copy, string
try:
    import sqlite3
except ImportError:
    logging.warn("No sqlite3 loaded")

from . import pubGeneric, pubConf, maxCommon, unicodeConvert, maxTables, pubPubmed

from os.path import *
from collections import namedtuple

# need to increase maximum size of fields for csv module
csv.field_size_limit(50000000)

# DATA FIELD DEFINITIONS

articleFields=[
"articleId",  # internal number that identifies this article in the pubtools system
"externalId", # original string id of the article, e.g. PMC12343 or the PPI or PMID123123
"source",  # the data format of the article, "elsevier", "medline", "pmc" or "crawler"
"publisher", # the name of the publisher icon, e.g. "pnas" or "aai", for all of PMC just 'pmc'
"origFile", # the original file where the article came from, e.g. the zipfile or genbank file
"journal",      # journal or book title
"printIssn",    # ISSN of the print edition of the article
"eIssn",        # optional: ISSN of the electronic edition of the journal/book of the article
"journalUniqueId", # medline only: NLM unique ID for journal
"year",         # first year of publication (electronic or print or advanced access)
"articleType", # research-article, review or other
"articleSection",  # elsevier: the section of the book/journal, e.g. "methods", "chapter 5" or "Comments"
"authors",  # list of author names, usually separated by semicolon
"authorEmails",  # email addresses of authors
"authorAffiliations",  # authors' affiliations
"keywords", # medline: mesh terms or similar, separated by / (medline is using , internally)
"title",    # title of article
"abstract", # abstract if available
"lang",     # article language
"vol",      # volume
"issue",    # issue
"page",            # first page of article, can be ix, x, or S4
"pmid",            # PubmedID if available
"pmidVersion",     # PubmedID version
"pmcId",           # Pubmed Central ID
"doi",             # DOI, without leading doi:
"pii",             # Publisher Item Identifier, only used by Elsevier and ACM
"fulltextUrl",     # URL to fulltext of article
"medlineCreatedDate",  # date medline entry was created
"medlineCompletedDate", # date medline entry was completed
"medlineRevisedDate", # date medline entry was last revised
"time",     # entry creation time (conversion time)
"offset",   # offset in .files, number of bytes (NOT number of unicode characters).
"size",      # total size (in bytes, not utf8 characters) of all files in this article + size of abstract
]

fileDataFields = [
"fileId", # numerical ID of the file: its article (ID * 1000)+some count (for suppl files)
"externalId", # copy of external ID, for quick greps and mallet
"articleId", # numerical ID of the article
"url", # the url where the file is located, can also be elsevier://, pmcftp:// etc
"desc", # a description of the file, e.g. main text, html title or supp file description
"fileType", # can be either "main" or "supp"
"time", # time/day of conversion from PDF/html/etc
"mimeType", # mimetype of original file before to-text-conversion
"locFname", # filename of original file before to-text-conversion
"content" # the data from this file (newline => \a, tab => space, cr => space, \m ==> \a)
]


refArtFields = ["articleId", "externalId", "artDoi", "artPmid"]
refFields    = ["authors", "title", "journal", "year", "month", "vol", "issue", "page", "pmid", "doi"]

ArticleRec = namedtuple("ArticleRec", articleFields)
emptyArticle = ArticleRec(*len(articleFields)*[""])

FileDataRec = namedtuple("FileRecord", fileDataFields)
emptyFileData = FileDataRec(*len(fileDataFields)*[""])

RefRec = namedtuple("citRec", refFields)

def isoGmtTime():
    "get current GMT time in ISO time, as it sorts correctly"
    return time.strftime("%Y-%m-%dT%T%z", time.gmtime())


def createEmptyFileDict(url=None, time=isoGmtTime(), mimeType=None, content=None,
    fileType=None, desc=None, externalId=None, locFname=None):
    fileData = emptyFileData._asdict()
    if time!=None:
        fileData["time"]=time
    if url!=None:
        fileData["url"]=url
    if mimeType!=None:
        fileData["mimeType"]=mimeType
    if externalId!=None:
        fileData["externalId"]=externalId
    if content!=None:
        fileData["content"]=content
    if fileType!=None:
        fileData["fileType"]=fileType
    if desc!=None:
        fileData["desc"]=desc
    if locFname!=None:
        fileData["locFname"] = locFname
    logging.log(5, "Creating new file record, url=%s, fileType=%s, desc=%s" % (url, fileType, desc))
    return fileData

def createEmptyArticleDict(pmcId=None, source=None, externalId=None, journal=None, \
    id=None, origFile=None, authors=None, fulltextUrl=None, keywords=None, title=None, abstract=None, \
    publisher=None, pmid=None, doi=None, lang=None):
    """ create a dictionary with all fields of the ArticleType """
    metaInfo = emptyArticle._asdict()
    metaInfo["time"]=isoGmtTime()
    if publisher!=None:
        metaInfo["publisher"]=publisher
    if doi:
        metaInfo["doi"]=doi
    if pmid:
        metaInfo["pmid"]=pmid
    if pmcId:
        metaInfo["pmcId"]=pmcId
    if origFile:
        metaInfo["origFile"]=origFile
    if source:
        metaInfo["source"]=source
    if journal:
        metaInfo["journal"]=journal
    if title:
        metaInfo["title"]=title
    if authors:
        metaInfo["authors"]=authors
    if fulltextUrl:
        metaInfo["fulltextUrl"]=fulltextUrl
    if keywords:
        metaInfo["keywords"]=keywords
    if externalId:
        metaInfo["externalId"]=externalId
    if abstract:
        metaInfo["abstract"]=abstract
    if lang:
        metaInfo["lang"]=lang
    return metaInfo

def splitTabFileOnChunkId(filename, outDir, chunkSize=None, chunkCount=None):
    """
    use the chunkId field of a tab-sep file as the output filename.
    if chunkSize is specified, ignore the chunkId field and make sure that each piece
    has chunkSize lines.
    """
    if isdir(outDir):
        logging.info("Deleting %s" % outDir)
        shutil.rmtree(outDir)

    if not os.path.isdir(outDir):
        logging.info("Creating directory %s" % outDir)
        os.makedirs(outDir)
    maxCommon.mustBeEmptyDir(outDir)

    # read data into data dict and split by "chunkId" field
    headerLine = open(filename).readline()
    logging.info("Reading %s, splitting into pieces" % filename)
    data = {}
    i = 0
    for row in maxCommon.iterTsvRows(filename, encoding=None):
        if chunkSize==None and chunkCount==None:
            chunkId = row.chunkId
        elif chunkSize!=None:
            chunkId = "%05d" % (i / chunkSize)
        elif chunkCount!=None:
            chunkId = "%05d" % (i % chunkSize)
        data.setdefault(str(chunkId), []).append("\t".join(row)+"\n")
        i += 1

    # write to outDir
    logging.info("Splitting file data, Writing to %d files in %s/xxxx.tgz" % (len(data), outDir))
    pm = maxCommon.ProgressMeter(len(data))
    for chunkIdString, lines in data.items():
        outfname = os.path.join(outDir, chunkIdString)
        logging.debug("Writing to %s" % outfname)
        fh = open(outfname, "w")
        fh.write(headerLine)
        for line in lines:
            fh.write(line)
        fh.close()
        pm.taskCompleted()

    return list(data.keys())

def toUnicode(var):
    """
    if string is not already a unicode strin (can happen due to upstream programming error):
    force variable to a unicode string, by decoding from utf8 first, then latin1 """
    if isinstance(var, str):
        return var
    elif type(var)==type(1):
        var = str(var)
    elif var==None:
        var = "NotSpecified"
    elif isinstance(var, str):
        try:
            var = var.decode("utf8")
        except UnicodeDecodeError as msg:
            logging.debug("Could not decode %s as utf8, error msg %s" % (var, msg))
            var = var.decode("latin1")
    return var

def listToUtf8Escape(list):
    """ convert list of variables to utf8 string as well as possible and replace \n and \t"""
    utf8List=[]
    for var in list:
        var = toUnicode(var)
        var = replaceSpecialChars(var)
        var = var.encode("utf8")
        utf8List.append(var)
    return utf8List

def dictToUtf8Escape(dict):
    """ convert dict of variables to utf8 string as well as possible and
    replace \n and \t
    """
    if dict==None:
        return None
    utf8Dict={}
    for key, var in dict.items():
        var = toUnicode(var)
        var = replaceSpecialChars(var)
        utf8Dict[key]=var
    return utf8Dict

def removeTabNl(var):
    " remove tab and all forms (!) of newline characters from string "
    # RAHHH! CRAZY UNICODE LINEBREAKS
    # the following would not work because of python's interpretation of unicode
    #newDict[key] = val.replace("\t", " ").replace("\n", " ")
    # there are more newlines than just \n and \m when one is using the
    # 'for line in file' construct in python
    # so we do this
    if type(var)==int:
        return str(var)
    cleanString = " ".join(var.splitlines()).replace("\t", " ")
    #logging.debug("cleaned string is %s" % repr(newStr))
    return cleanString

def articleDictToTuple(artDict):
    " convert a dict to an article tuple "
    for key in articleFields:
        if key not in artDict:
            artDict[key] = ""
    artTuple = ArticleRec(**artDict)
    return artTuple

class PubWriterFile:
    """
    a class that stores article and file data into tab-sep files
    that are located in a subdirectory.

    Constructor will create two files:
        <chunkId>.files    = raw file content and some meta (filetype, url)
        <chunkId>.articles = article meta data (author, year, etc)

    The constructor takes only the .files or .articles base filename as input and will
    derive outDir and chunkId from it.

    all writes will first go to tempDir, and will only be copied over
    to outDir on .close()
    We only copy files over if any files data was actually written
    We only copy articles over if any articles data was actually written

    """
    def __init__(self, fileDataFilename):
        self.articlesWritten = 0
        self.filesWritten = 0
        self.tempDir = pubConf.getTempDir()

        # convoluted way to find output filenames
        # needed because of parasol
        outDir = os.path.dirname(fileDataFilename)
        fileDataBasename = os.path.basename(fileDataFilename)
        chunkId = fileDataBasename.split(".")[0]

        ext = ".gz" # use .gz extension even if not compressing, saves a ton of file name problems
        if pubConf.compress:
            if bgzfLoaded:
                openFunc = Bio.bgzf.open
            else:
                openFunc = gzip.open
        else:
            openFunc = open

        self.fileBaseName = chunkId+".files"+ext
        articleBaseName = chunkId+".articles"+ext
        refBaseName = chunkId+".refs"

        # setup reference table handle
        self.refDir = join(outDir, "refs")
        self.tempRefName = join(self.tempDir, refBaseName)
        self.refFh = None

        self.finalArticleName = join(outDir, articleBaseName)
        self.finalFileDataName    = join(outDir, self.fileBaseName)
        self.finalRefFname = join(self.refDir, refBaseName)

        # setup file and article table handles
        # in temporary directory, so we do not leave behind half-written
        # chunks. Temp files are moved over to final on self.close()
        #self.tmpFileFname = os.path.join(self.tempDir, self.fileBaseName)
        self.tmpFileFname = tempfile.mktemp(prefix="pubStore.files.")
        self.fileFh = openFunc(self.tmpFileFname, "w")
        self.fileFh.write("#"+"\t".join(fileDataFields)+"\n")

        #self.tmpArticleFname = os.path.join(self.tempDir, articleBaseName)
        self.tmpArticleFname = tempfile.mktemp(prefix="pubStore.article.")
        self.articleFh = openFunc(self.tmpArticleFname, "w")
        self.articleFh.write("#"+"\t".join(articleFields)+"\n")

        maxCommon.delOnExit(self.tmpFileFname)
        maxCommon.delOnExit(self.tmpArticleFname)

        self.outFilename = os.path.join(outDir, fileDataBasename)
        logging.debug("pubStore writer open. tmp files %s and %s. Dest files %s and %s" % \
            (self.tmpArticleFname, self.tmpFileFname, self.finalArticleName, self.finalFileDataName))

    def _removeSpecChar(self, lineDict):
        " remove tab and NL chars from values of dict "
        newDict = {}
        for key, val in lineDict.items():
            newDict[key] = removeTabNl(val)
        return newDict

    def writeRefs(self, artDict, refRows):
        " write references to table in refs/ subdir"
        if self.refFh==None:
            # lazily open ref file, add headers
            if not os.path.isdir(self.refDir):
                try:
                    os.makedirs(self.refDir)
                except OSError:
                    logging.info("makedir %s failed, probably just race condition" % self.refDir)
                    pass
            self.refFh = open(self.tempRefName, "w")
            logging.info("Created tempfile for refs %s" % self.refFh.name)
            maxCommon.delOnExit(self.tempRefName)

            refHeaders = copy.copy(refArtFields)
            refHeaders.extend(refFields)
            self.refFh.write("#"+"\t".join(refHeaders)+"\n")

        # prepare a list of article IDs of the source article
        srcArtFields = []
        for artField in refArtFields:
            if artField=="artDoi":
                artField="doi"
            if artField=="artPmid":
                artField="pmid"
            artVal = artDict[artField]
            srcArtFields.append(artVal)
        srcPrefix = "\t".join(srcArtFields)+"\t"

        # output all references
        logging.debug("Writing %d references for article %s" % (len(refRows), artDict["externalId"]))
        for ref in refRows:
            # output the source article IDs
            self.refFh.write(srcPrefix.encode("utf8"))

            # output the reference article fields
            self.refFh.write('\t'.join(ref).encode("utf8"))
            self.refFh.write("\n")

    def writeFile(self, articleId, fileId, fileDict, externalId=""):
        """ appends id and data to current .file table,
        """
        assert("content" in fileDict)

        if fileDict["content"]==None:
            logging.warn("file %s, object or content is None" % fileId)
            fileDict["content"] = ""

        if "externalId" not in fileDict:
            fileDict["externalId"] = fileDict["articleId"]

        self._checkFields(fileDict, fileDataFields)

        if "externalId" not in fileDict:
            fileDict["externalId"] = externalId
        fileDict["fileId"]=str(fileId)
        fileDict["articleId"]=str(articleId)
        # convert dict to line and write to xxxx.file
        fileTuple = FileDataRec(**fileDict)
        fileTuple = listToUtf8Escape(fileTuple)
        line = "\t".join(fileTuple)

        #logging.log(5, "Writing line to file table, dict is %s" % fileDict)
        self.filesWritten += 1
        self.fileFh.write(line+"\n")

    def _checkFields(self, inDict, fieldList):
        """ makes sure that inDict contains one key for each string in fieldList.
        Keys not found are set to empty strings.
        Aborts if a key is found in inDict that is not in fieldList.
        """
        gotFields = set(inDict.keys())
        mustHaveFields = set(fieldList)
        unknownFields = gotFields - mustHaveFields
        if len(unknownFields)>0:
            raise Exception("article/file writer got unknown fields: %s" % repr(unknownFields))

        missingFields = mustHaveFields - gotFields
        logging.log(5, "PubStore Writer: Adding empty strings for: %s" % missingFields)
        for key in missingFields:
            inDict[key] = ""
        return inDict


    def writeArticle(self, articleId, articleDict):
        """ appends data to current chunk. article info has to be written before the files,
        otherwise the offset in the article dict will be wrong.
        """
        logging.log(5, "appending article info to %s: %s" % (self.tmpArticleFname, str(articleDict)))
        articleDict["articleId"]=articleId

        # fill the "offset" field
        filePos = 0
        if self.fileFh is not None:
            filePos = self.fileFh.tell()
        articleDict["offset"] = str(filePos)

        articleDict = self._removeSpecChar(articleDict)
        articleDict = self._checkFields(articleDict, articleFields)

        articleTuple = ArticleRec(**articleDict)

        # convert all fields to utf8 string, remove \n and \t
        articleTuple = listToUtf8Escape(articleTuple)

        line = "\t".join(articleTuple)
        self.articleFh.write(line+"\n")
        self.articlesWritten += 1
        logging.log(5, "%d articles written" % self.articlesWritten)

    def writeDocs(self, artDict, fileDicts):
        " write all article and files in one go. Optionally extract images from PDFs. "
        # set the 'size' field of the article
        totalSize = 0
        for fileDict in fileDicts:
            totalSize += len(fileDict['content'])
        totalSize += len(artDict['abstract'])
        artDict['size'] = str(totalSize)

        self.writeArticle(artDict['articleId'], artDict)

        for fileDict in fileDicts:
            self.writeFile(artDict['articleId'], fileDict['fileId'], fileDict)


    def close(self, keepEmpty=False):
        """
        close the 3 files, move them over to final targets
        """
        logging.debug("Moving local tempfiles over to files on server %s" % self.finalArticleName)

        self.fileFh.close()
        self.articleFh.close()

        if self.articlesWritten==0:
            logging.warn("No articles received, not writing anything, but creating a 0 sized file for parasol")
            # just create a 0-size file for parasol
            open(self.finalArticleName, "w")

        if self.filesWritten > 0 or keepEmpty:
            logging.debug("moving articles table to %s" % self.finalArticleName)
            shutil.move(self.tmpFileFname, self.finalFileDataName)

        if self.articlesWritten > 0 or keepEmpty:
            logging.debug("moving files table to %s" % self.finalFileDataName)
            shutil.move(self.tmpArticleFname, self.finalArticleName)

        if self.refFh!=None:
            self.refFh.close()
            shutil.move(self.refFh.name, self.finalRefFname)

def createPseudoFile(articleData):
    """ create a file from the abstract and title of an article,
    for articles that don't have fulltext (pubmed) """
    logging.debug("no file data, creating pseudo-file from abstract")
    fileData = createEmptyFileDict()
    fileData["url"] = articleData.fulltextUrl
    fileData["content"] = " ".join([" ",articleData.title, articleData.abstract, " "])
    fileData["mimeType"] = "text/plain"
    fileData["fileId"] = int(articleData.articleId) * (10**pubConf.FILEDIGITS)
    fileTuple = FileDataRec(**fileData)
    return fileTuple

class PubReaderFile:
    """
    read articles from tab-sep files
    """
    def __init__(self, fname):
        " fname can end in .articles.gz, reader will still read both articles and files "
        logging.debug("Reading data from file with prefix %s (.articles.gz, .files.gz)" % fname)
        baseDir = dirname(fname)
        base = basename(fname).split('.')[0]
        articleFn = join(baseDir, base+".articles.gz")
        fileFn = join(baseDir, base+".files.gz")
        logging.debug("Reading %s and %s" % (articleFn, fileFn))

        self.articleRows = None
        if isfile(articleFn) and getsize(articleFn)!=0:
            self.articleRows = maxCommon.iterTsvRows(articleFn, encoding="utf8")

        self.fileRows = None
        if isfile(fileFn) and getsize(fileFn)!=0:
            self.fileRows  = maxCommon.iterTsvRows(fileFn, encoding="utf8")

    def iterFileRows(self):
        """ iterate over file data """
        return self.fileRows

    def iterArticleRows(self):
        """ iterate over article data """
        return self.articleRows

    def _readFilesForArticle(self, articleId, fileDataList):
        " reads files until the articleId changes, adds them to fileDataList "

        for fileData in self.fileRows:
           logging.log(5, "Read file data %s for article %s" % \
               (str(fileData.fileId), fileData.articleId))
           text = pubGeneric.forceToUnicode(fileData.content)
           fileData = fileData._replace(content=text)
           if articleId==fileData.articleId:
               logging.log(5, "adding file data")
               fileDataList.append(fileData)
           else:
               fileIds = list(set([str(x.fileId)[:pubConf.ARTICLEDIGITS] for x in fileDataList]))
               logging.log(5, "article change. yielding: articleId %s, %d files with ids %s" % \
                   (articleId, len(fileDataList), fileIds))
               assert(len(fileIds)==1)
               assert(fileIds[0]==str(articleId))
               return fileDataList, fileData
        return fileDataList, None

    def _keepOnlyMain(self, files):
        " remove all suppl files "
        mainFiles = {}
        newFiles = []
        for fileData in files:
            if fileData.fileType=="main" or fileData.fileType=="":
                newFiles.append(fileData)
        logging.log(5, "Main-text filter: got %d files, returned %d files" % (len(files), len(newFiles)))
        return newFiles

    def _keepBestMain(self, files, preferType):
        """ if there are several main text formats, keep only one of them.
        preferType can be "pdf" or "xml" (which includes html files)
        Keep all other files
        """
        mainFiles = {}
        newFiles = []
        for fileData in files:
            if fileData.fileType=="main" or fileData.fileType=="":
                # we should never have two main files with same type
                assert(fileData.mimeType not in mainFiles)
                mainFiles[fileData.mimeType] = fileData
            else:
                newFiles.append(fileData)

        # this should happen only very very rarely, if main file was corrupted
        if len(mainFiles)==0:
            logging.error("No main file for article?")
            logging.error("%s" % files)
            return newFiles

        if len(mainFiles)==1:
            newFiles.insert(0, list(mainFiles.values())[0])
            return newFiles

        # remove the pdf if there are better files
        if preferType=="xml":
            if "application/pdf" in mainFiles and \
                    ("text/xml" in mainFiles or "text/html" in mainFiles):
                logging.debug("Removing pdf")
                del mainFiles["application/pdf"]

        elif preferType=="pdf":
            # remove the xml if there are PDF files
            if "application/pdf" in mainFiles:
                if "text/xml" in mainFiles:
                    del mainFiles["text/xml"]
                    logging.debug("Removing xml")
                if "text/html" in mainFiles:
                    del mainFiles["text/html"]
                    logging.debug("Removing html")
        else:
            assert(False)

        # paranoia check: make sure that we still have left one file
        if not len(mainFiles)>=1:
            logging.error("no main file: in %s out %s " % (files, mainFiles))
            raise Exception("no main file left")

        newFiles.insert(0, list(mainFiles.values())[0])
        return newFiles

    def iterArticlesFileList(self, algPrefs):
        """ iterate over articles AND files, as far as possible

        for input files with article and file data:
            yield a tuple (articleData, list of fileData) per article
        for input files with no article data, yield a tuple (None, [fileData])
        for input files with no file data, generate pseudo-file from abstract
        (title+abstract+authors)

        algPrefs.onlyMeta == True: do not read .files.gz and yield
        (articleData, pseudoFile)
        algPrefs.onlyMain == True: skip supplemental files
        algPrefs.preferXml == True: run on only one main text file.
        skip the PDF file if there are PDF and XML/Html main files.
        algPrefs.preferPdf == True: run on only one main text file.
        skip XML or HTML files if there are multiple file types.
        """
        fileDataList = []
        lastFileData = None
        if self.articleRows==None:
            raise StopIteration

        for articleData in self.articleRows:
            if len(fileDataList)!=0 and fileDataList[0].articleId!=articleData.articleId:
                logging.warn("skipping %s, seems that articleId is out of sync with files" %
                    articleData.articleId)
                continue
            logging.log(5, "Read article meta info for %s" % str(articleData.articleId))

            if self.fileRows!=None and (algPrefs is None or (algPrefs!=None and algPrefs.onlyMeta==False)):
                # if file data is there and we want it, read as much as we can
                fileDataList, lastFileData = \
                    self._readFilesForArticle(articleData.articleId, fileDataList)

                if algPrefs!=None:
                    if algPrefs.onlyMain:
                        fileDataList = self._keepOnlyMain(fileDataList)
                    if algPrefs.preferXml:
                        fileDataList = self._keepBestMain(fileDataList, "xml")
                    if algPrefs.preferPdf:
                        fileDataList = self._keepBestMain(fileDataList, "pdf")

                yield articleData, fileDataList
                fileDataList = [lastFileData]

            else:
                # if only abstract (e.g. medline): create pseudo file
                fileTuple = createPseudoFile(articleData)
                yield articleData, [fileTuple]

        if len(fileDataList)!=0 and lastFileData!=None:
            logging.log(5, "last line: yielding last article + rest of fileDataList")
            yield articleData, fileDataList

    def iterFileArticles(self):
        """ iterate over files and also give the articleData for each file """
        self.articleIdToData = {}
        # normal operation, if we have files
        if self.fileRows!=None:
            for articleData in self.articleRows:
                self.articleIdToData[articleData.articleId] = articleData

            for fileData in self.fileRows:
               articleData = self.articleIdToData.get(fileData.articleId, None)
               yield fileData, articleData
        # just return articles and blank file if we don't have files
        else:
            for articleData in self.articleRows:
                yield emptyFileData, articleData

    def iterFiles(self):
        """ compat to iterFileArticles, yields only Files """
        for fileData in self.fileRows:
           yield fileData, None

    def iterArticles(self):
        """ compat to iterFileArticles, yields only articles """
        for articleData in self.articleRows:
           yield None, articleData

    def close(self):
        if self.articleRows:
            self.articleRows.close()
        if self.fileRows:
            self.fileRows.close()

#class PubReaderPmidDir:
#    """ reads files from a directory of text files in format <pmid>.txt"""
#    def __init__(self, dirName):
#        self.dirName = dirName
#
#    def iterArticlesFileList(self, onlyMeta=False, onlyBestMain=False, onlyMain=False):
#        class ArtData:
#            def _replace(self, content=None):
#                return self
#
#            def _asdict(self):
#                return {"pmid":10, "externalId":"extId000", "articleId":100000000}
#
#
#        for fname in glob.glob(join(self.dirName, "*.txt")):
#            pmid = splitext(basename(fname))[0]
#            art = ArtData()
#            art.articleId = pmid
#            art.pmid = pmid
#            art.externalId =
#            art.printIssn = "1234-1234"
#
#            fileObj = C()
#            fileObj.fileId = "1001"
#            fileObj.content = self.text
#            fileObj.fileType = "main"
#
#        yield art, [fileObj]

class PubReaderTest:
    """ reads a single text file or a directory of text files"""
    def __init__(self, fname, text=None):
        if fname!=None:
            self.text = None
            if isdir(fname):
                self.fnames = glob.glob(join(fname, "*.txt"))
            elif isfile(fname):
                self.fnames = [fname]
            else:
                raise Exception("Could not open %s" % fname)
        elif text!=None:
            self.text=text
        else:
            assert(False)
        self.artId = 1000000000


    def _makeArtFile(self, text, fname):
        class C:
            def _replace(self, content=None):
                return self

            def _asdict(self):
                return {"pmid":10, "externalId":"extId000", "articleId":100000000}

        art = C()
        art.articleId = str(self.artId)
        self.artId += 1
        fnameBase = basename(fname.replace("PMID","").replace("PMC","")).split(".")[0]
        if fnameBase.isdigit():
            art.pmid = fnameBase
        else:
            art.pmid = "000000"
        art.externalId = fname
        art.printIssn = "1234-1234"
        art.url =  "http://www.pubmed.com"

        fileObj = C()
        fileObj.fileId = "1000"
        fileObj.content = text
        fileObj.fileType = "main"
        fileObj.externalId = "file0000"
        fileObj.desc="desc"

        lines = text.splitlines()
        abstract = ""
        title = ""
        for line in lines:
            if line.startswith("abstract: "):
                abstract = string.split(line, ": ", 1)[1]
            if line.startswith("title: "):
                title = string.split(line, ": ", 1)[1]
        art.title = title
        art.abstract = abstract

        return art, [fileObj]

    def iterArticlesFileList(self, algPrefs):
        if self.text != None:
            yield self._makeArtFile(self.text, "1000")

        for fname in self.fnames:
            text = open(fname).read()
            yield self._makeArtFile(text, fname)

def iterArticleDirList(textDir, algPrefs=None):
    """ iterate over all files with article/fileData in textDir.
    This yields tuples (articleDict, list of fileDicts)
    """
    logging.debug("Getting filenames in dir %s" % textDir)
    fileNames = getAllArticleFnames(textDir)
    logging.debug("Found %d files in input dir %s" % (len(fileNames), textDir))
    pm = maxCommon.ProgressMeter(len(fileNames))
    for textCount, textFname in enumerate(fileNames):
        reader = PubReaderFile(textFname)
        logging.debug("Reading %s, %d files left" % (textFname, len(fileNames)-textCount))
        pr = PubReaderFile(textFname)
        artIter = pr.iterArticlesFileList(algPrefs)
        for article, fileList in artIter:
            yield article, fileList
        pm.taskCompleted()

def iterArticleDataDirs(textDirs, type="articles", filterFname=None, updateIds=None):
    logging.info("Getting rows from %s files in dirs: %s" % (type, textDirs))
    for textDir in textDirs:
        for row in iterArticleDataDir(textDir, type, filterFname, updateIds):
            yield row

def iterArticleDataDir(textDir, type="articles", filterFname=None, updateIds=None):
    """ yields all articleData from all files in textDir
        Can filter to yield only a set of filenames or files for a
        given list of updateIds.
    """
    fcount = 0
    assert(type in ["articles", "files"])

    if isfile(textDir):
        fileNames = [textDir]
        logging.debug("Found 1 file, %s" % textDir)
    else:
        fileNames = getAllArticleFnames(textDir, type=type)
        logging.debug("Looking for all fulltext files in %s, found %d files" % \
            (textDir, len(fileNames)))
        if updateIds!=None and len(updateIds)!=0:
            logging.debug("Restricting fulltext files to updateIds %s" % str(updateIds))
            filteredFiles = []
            for updateId in updateIds:
                for fname in fileNames:
                    if basename(fname).startswith(str(updateId)+"_"):
                        filteredFiles.append(fname)
                logging.debug("Update Id %s, %d files" % (str(updateId), len(filteredFiles)))
            fileNames = list(filteredFiles)
        logging.debug("Found %d files in input dir %s" % (len(fileNames), textDir))

    pm = maxCommon.ProgressMeter(len(fileNames), stepCount=100)
    for textFname in fileNames:
        if filterFname!=None and not filterFname in textFname:
            logging.warn("Skipping %s, because file filter is set" % textFname)
            continue
        reader = PubReaderFile(textFname)
        logging.debug("Reading %s, %d files left" % (textFname, len(fileNames)-fcount))
        fcount+=1
        if type=="articles":
            for articleData in reader.articleRows:
                if "publisher" not in articleData._fields: # XX temporary bugfix as I have some old files
                    articleData = list(articleData)
                    articleData.insert(2, "")
                    articleData[3] = ""
                yield articleData
        elif type=="files":
            for fileData in reader.fileRows:
                yield fileData
        elif type=="annots":
            for row in maxCommon.iterTsvRows(textFname):
                yield row
        else:
            assert(False) # illegal type parameter
        pm.taskCompleted()

all_chars = (chr(i) for i in range(0x110000))
control_chars = ''.join(map(chr, list(range(0,7)) + list(range(8,32)) + list(range(127,160))))
control_char_re = re.compile('[%s]' % re.escape(control_chars))

def replaceSpecialChars(string):
    " replace all special characters with space and linebreaks to \a = ASCII 7 = BELL"
    if string==None:
        return ""
    string = string.replace('\u2028', '\n') # one of the crazy unicode linebreaks
    string = "\n".join(string.splitlines()) # get rid of other crazy unicode linebreaks
    string = string.replace("\m", "\a") # old mac text files
    string = string.replace("\n", "\a")
    string = control_char_re.sub(' ', string)
    return string

space_re = re.compile('[ ]+')

def prepSqlString(string, maxLen=pubConf.maxColLen):
    """ change <<</>>> to <b>/</b>, replace unicode chars with
    character code, because genome browser html cannot do unicode

    """
    global control_chars
    if string==None:
       string = ""
    string = toUnicode(string)
    string = replaceSpecialChars(string)
    string = string.replace("\\", "\\\\") # mysql treats \ as escape char on LOAD DATA
    string = string.replace("<<<", "<B>")
    string = string.replace(">>>", "</B>")
    string = string.replace("\A", "<BR>")
    string = space_re.sub(' ', string)
    string = unicodeConvert.string_to_ncr(string)
    if len(string) > maxLen:
       logging.warn("Cutting column to %d chars, text: %s" % (maxLen, string[:200]))
       string = string[:maxLen]
    return string

#def iterFileDataDir(textDir):
    #return iterArticleDataDir(textDir, type="files")

def getAllBatchIds(outDir):
    """ parse batches.tab and return all available batchIds
    """
    batchIds = []
    for row in maxCommon.iterTsvRows(join(outDir, "batches.tab")):
        batchIds.append(row.batchId)
    logging.debug("Found batchIds %s in directory %s" % (batchIds, outDir))
    return batchIds

def parseUpdatesTab(outDir, minArticleId):
    """ parse updates.tab and find next available articleIds and
    list of files that were processed in all updates

    returns (next free updateId, next free articleId, list of files that have been processed)

    """
    inFname = join(outDir, "updates.tab")

    if not isfile(inFname):
        logging.info("could not find %s, this seems to be the first run in this dir" % inFname)
        return 0, minArticleId, []
    logging.info("Reading IDs of files that are already done from %s" % inFname)

    doneFiles = set()
    row = None
    logging.debug("Parsing %s" % inFname)
    for row in maxCommon.iterTsvRows(inFname, encoding="utf8"):
        rowFiles = row.files.split("|")
        doneFiles.update(rowFiles)

    if row==None:
        logging.warn("empty file %s, this seems to be the first run in this dir" % inFname)
        return 0, minArticleId, []
    logging.debug("Parsed updates.tab, %d files already done" % (len(doneFiles)))
    return int(row.updateId)+1, int(row.lastArticleId)+1, doneFiles

def listAllUpdateIds(textDir):
    " return set of possible update Ids in textDir "
    logging.info("Getting list of available update ids in text directory")
    inFname = join(textDir, "updates.tab")
    updateIds = set()
    if not isfile(inFname):
        logging.info("Could not find %s, using filenames to get updates" % inFname)
        inNames = glob.glob(join(textDir, "*.articles.gz"))
        updateIds = set([basename(x).split("_")[0] for x in inNames])
        logging.info("Found text update ids from filenames: %s" % updateIds)
        return updateIds

    for row in maxTables.TableParser(inFname).lines():
        updateIds.add(row.updateId)
    logging.info("Text update ids in %s: %s" % (inFname, updateIds))
    return updateIds

def guessChunkSize(outDir):
    " get line count of  0_00000.articles.gz in outDir"
    fname = join(outDir, "0_00000.articles.gz")
    if not isfile(fname):
        fname = join(outDir, "0_00000.articles")
        if not isfile(fname):
            return None
        lineCount = len(open(fname).readlines())
    else:
        lineCount = len(gzip.open(fname).readlines())
    logging.info("Guessing chunk size: Chunk size of %s is %d" % (fname, lineCount))
    return lineCount

def appendToUpdatesTxt(outDir, updateId, maxArticleId, files):
    " append a line to updates.tab in outDir, create file if necessary "
    outFname = join(outDir, "updates.tab")
    if len(files)==0:
        logging.info("Not writing any progress update, no new files")
        return

    logging.info("Writing progress to %s: updateId %d, %d files" % (outFname, updateId, len(files)))
    #logging.info("You must delete the last line of %s if the cluster job fails" % outFname)
    if not isfile(outFname):
        outFh = open(outFname, "w")
        headers = ["updateId", "lastArticleId", "time", "files"]
        outFh.write("\t".join(headers))
        outFh.write("\n")
    else:
        outFh = open(outFname, "a")

    row = [str(updateId), str(maxArticleId), isoGmtTime(), "|".join(files)]
    outFh.write("\t".join(row))
    outFh.write("\n")
    outFh.close()

def moveFiles(srcDir, trgDir, nameList=None):
    """ move all files from src to target dir, prefix is typically sth like ["articles", "files"] """
    logging.info("Moving files from %s to %s" % (srcDir, trgDir))
    if not isdir(trgDir):
        logging.info("Creating directory %s" % trgDir)
        os.makedirs(trgDir)

    for fname in os.listdir(srcDir):
        if nameList and not basename(fname).split(".")[1] in nameList:
            logging.log(5, "Not moving %s" % fname)
            continue
        infname = join(srcDir, fname)
        outfname = join(trgDir, fname)
        #if isdir(infname) and not basename(infname) in subDirs:
            #continue
        if isdir(infname):
            continue
        if isfile(outfname):
            logging.debug("Deleting %s" % outfname)
            os.remove(outfname)
        logging.debug("moving %s to %s" % (infname, outfname))
        shutil.move(infname, outfname)

def articleIdToDataset(articleId):
    """ uses the central namespace table to resolve an articleId to its dataset name
    >>> articleIdToDataset(4300000011)
    'imgt'
    >>> articleIdToDataset(2004499279)
    'elsevier'
    """
    articleId = int(articleId)
    restList = []
    for datasetName, artIdStart in pubConf.identifierStart.items():
        rest = articleId - artIdStart
        if rest>0:
            restList.append( (datasetName, rest) )
    restList.sort(key=operator.itemgetter(1))
    return restList[0][0]

def iterChunks(datasets):
    """ given a list of datasets like ["pmc", "elsevier"], return a list of directory/chunkStems,
    e.g. "/hive/data/inside/pubs/text/elsevier/0_0000.articles.gz". Used to prepare cluster jobs

    If dataset is already a valid filename, will only return the filename (for debugging)
    """
    for dataset in datasets:
        if isfile(dataset):
            yield dataset
        else:
            dirName = join(pubConf.textBaseDir, dataset)
            for fname in getAllArticleFnames(dirName):
                yield fname

def addLoadedFiles(con, cur, fileNames):
    " given a sqlite db, create a table loadedFiles and add fileNames to it "
    #dbFname = getArtDbPath(inDir)
    #assert(isfile(dbFname))
    #fileNames = [(basename(x), ) for x in fileNames] # sqlite only accepts tuples, strip path
    #con, cur = maxTables.openSqlite(dbFname, lockDb=True)
    fileNames = [(basename(s),) for s in fileNames]
    cur.execute("CREATE TABLE IF NOT EXISTS loadedFiles (fname TEXT PRIMARY KEY);")
    con.commit()
    logging.debug("INSERTing %d filenames into table loadedFiles" % len(fileNames))
    sql = "INSERT INTO loadedFiles (fname) VALUES (?)"
    cur.executemany(sql, fileNames)
    con.commit()

def getUnloadedFnames(con, cur, newFnames):
    """ given a sqlite db and a list of filenames, return those that have not
    been loaded yet into the db. Looks only at basename of files.
    """
    loadedFnames = []
    try:
        for row in cur.execute("SELECT fname from loadedFiles"):
            loadedFnames.append(basename(row[0]))
    except sqlite3.OperationalError:
        logging.debug("No loadedFiles table yet")
        return newFnames
    logging.debug("Files that have been loaded already: %s" % loadedFnames)

    # keep only filenames that haven't been loaded yet
    loadedFnames = set(loadedFnames)
    toLoadFnames = []
    for newFname in newFnames:
        if basename(newFname) not in loadedFnames:
            toLoadFnames.append(basename(newFname))

    logging.debug("Files that have not been loaded yet: %s" % toLoadFnames)
    return toLoadFnames

def sortPubFnames(fnames):
    """
    sort names like 0_00000, 11_1111.articles.gz in the right order from 0_00000 to 1111_11111
    >>> sortPubFnames(["/hive/data/test/11_0000.articles.gz", "1_0000.articles.gz"])
    ['1_0000.articles.gz', '/hive/data/test/11_0000.articles.gz']
    """
    newList = []
    for fname in fnames:
        base = basename(fname).split(".")[0]
        update, chunk = base.split("_")
        update, chunk = int(update), int(chunk)
        sortKey = update*10000 + chunk
        newList.append((sortKey, fname))

    newList.sort(key=operator.itemgetter(0))
    newFnames = [el[1] for el in newList]
    return newFnames

def chunkIdFromFname(fname):
    " given the path of chunk, return it's chunk ID, like 0_00001 "
    return basename(fname).split(".")[0]

def addToDatabase(con, cur, tsvFnames):
    " load articles files into sqlLite db table. Adds a field 'chunkId'. "
    tableName = "articles"
    allFields = list(tuple(articleFields)) # make a deep copy of list
    allFields.append("chunkId")

    idxFields = ["pmid", "pmcId","printIssn", "eIssn", "year", "doi", "extId"]
    intFields = ["pmid", "pmcId","year","offset"]
    primKey = "articleId"

    # create table
    tableFields = articleFields
    tableFields.append("chunkId")
    createSql, idxSqls = maxTables.makeTableCreateStatement(tableName, articleFields, \
        intFields=intFields, idxFields=idxFields, primKey=primKey)
    logging.log(5, "creating table with %s" % createSql)
    cur.execute(createSql)
    con.commit()

    logging.info("Loading data into table")
    tp = maxCommon.ProgressMeter(len(tsvFnames))
    sql = "INSERT INTO %s (%s) VALUES (%s)" % (tableName, ", ".join(allFields), ", ".join(["?"]*len(allFields)))
    rowCount = 0
    for tsvName in tsvFnames:
        logging.debug("Loading file %s" % tsvName)
        if os.path.getsize(tsvName)==0:
            logging.debug("Skipping %s, zero size" % tsvName)
            continue
        rows = list(maxCommon.iterTsvRows(tsvName))
        chunkId = chunkIdFromFname(tsvName)

        # add the chunkId field
        newRows = []
        for row in rows:
            newRow = list(row)
            newRow.append(chunkId)
            newRows.append(newRow)
        rowCount += len(newRows)

        logging.log(5, "Running Sql %s against %d rows" % (sql, len(newRows)))
        cur.executemany(sql, newRows)
        con.commit()
        tp.taskCompleted()

    logging.info("Adding indexes to table")
    for idxSql in idxSqls:
        cur.execute(idxSql)
        con.commit()

    addLoadedFiles(con, cur, tsvFnames)
    logging.info("Loaded %s chunks into index, %d new rows" % (len(tsvFnames), rowCount))

#def loadNewTsvFilesSqlite(inDir, tableName, tsvFnames):
    #" load pubDoc files that are not loaded yet into sqlite db table, mark them as loaded at the end "
    #if len(tsvFnames)==0:
        #return
    #logging.debug("Preparing to load %d files into table %s, db dir %s" %(len(tsvFnames), tableName, inDir))
    #firstFname = tsvFnames[0]
    #if firstFname.endswith(".gz"):
        #firstFh = gzip.open(firstFname)
    #else:
        #firstFh = open(firstFname)
    #logging.debug("fields in first input file are: %s" % )

    #con, cur = openArticleDb(textDir)
    #toLoadFnames = getUnloadedFnames(con, cur, tsvFnames)
    #logging.debug("Loading %d files" % (len(toLoadFnames)))
    #toLoadFnames = sortPubFnames(toLoadFnames)
#
    #if len(toLoadFnames)==0:
        #logging.debug("No new files to load")
    #else:
        ##maxTables.loadTsvSqlite(dbFname, tableName, toLoadFnames, headers=headers, \
            #primKey="articleId", intFields=intFields, idxFields=indexedFields, dropTable=False)
        #con, cur = openArticleDb(inDir)
        #loadIndexes(con, cur, toLoadFnames)

def updateSqlite(textDir):
    """ load all .articles files that are not currently indexed
    into the sqlite database
    """
    artFnames = getAllArticleFnames(textDir)
    assert(len(artFnames)!=0) # there are no input files in the text data directory
    dbPath = getArtDbPath(textDir)
    if isfile(dbPath):
        con, cur = openArticleDb(textDir)
        copyBack = False
    else:
        # if db does not exist yet, use max speed: write to ramdisk and copy back.
        ramDbPath = pubGeneric.getFastUniqueTempFname()
        logging.info("First load. Creating temporary sqlite db on ramdisk %s" % ramDbPath)
        copyBack = True
        con, cur = maxTables.openSqlite(ramDbPath, lockDb=True)

    artFnames = [basename(x) for x in artFnames]
    toLoadFnames = getUnloadedFnames(con, cur, artFnames)
    toLoadPaths = [join(textDir, fname) for fname in toLoadFnames]

    addToDatabase(con, cur, toLoadPaths)
    con.close()

    if copyBack:
        shutil.copy(ramDbPath, dbPath)
        os.remove(ramDbPath)

datasetRanges = None

def setupDatasetRanges():
    global datasetRanges
    if datasetRanges!=None:
        return

    datasetStarts = list(pubConf.identifierStart.items())
    datasetStarts.sort(key=operator.itemgetter(1))
    assert(datasetStarts[-1][0]=="free")

    datasetRanges = []
    for i in range(0, len(datasetStarts)-1):
        d = datasetStarts[i]
        d2 = datasetStarts[i+1]
        datasetRanges.append((d[0], d[1], d2[1]))

def artIdToDatasetName(artId):
    """ resolve article Id to dataset name
    >>> artIdToDatasetName(1000000000)
    'pmc'
    >>> artIdToDatasetName(4500000005)
    'yif'
    """
    artId = int(artId)
    setupDatasetRanges()
    for dataset, minId, maxId in datasetRanges:
        if artId >= minId and artId < maxId:
            return dataset

def getArtDbPath(datasetName):
    """ return the sqlite database name with meta info of a dataset """
    dataDir = pubConf.resolveTextDir(datasetName, mustFind=True)
    dbPath = join(dataDir, "articles.db")
    return dbPath

conCache = {}

def openArticleDb(datasetName, mustOpen=False, useRamdisk=False):
    " open an article sqlite DB, return (conn, cur) tuple "
    if datasetName in conCache:
        con, cur = conCache[datasetName]
    else:
        path = getArtDbPath(datasetName)
        if path is None or not isfile(path):
            logging.debug("Creating new file %s" % path)
            if mustOpen:
                raise Exception("Could not open dataset %s" % datasetName)
            #return None, None
        if useRamdisk:
                ramDbPath = pubGeneric.getFastUniqueTempFname()
                logging.info("Copying %s to %s for faster access" % (path, ramDbPath))
                maxCommon.delOnExit(ramDbPath)
                shutil.copy(path, ramDbPath)
                path = ramDbPath

        logging.debug("Opening db %s" % path)
        con, cur = maxTables.openSqlite(path, asDict=True)
        conCache[datasetName] = (con,cur)
    return con, cur

def lookupArticleByArtId(artId):
    " convenience method to get article info given article Id, caches db connections "
    dataset = artIdToDatasetName(artId)
    con, cur = openArticleDb(dataset)
    return lookupArticle(con, cur, "articleId", artId)

connCache = {}

def lookupArticleByPmid(datasets, pmid, preferLocal=True):
    """ convenience method to get article info given pubmed Id, caches db connections.
        Uses eutils if local medline is not available and preferLocal is False
    """
    for dataset in datasets:
        con = None
        if preferLocal:
            # keep cache of db connections
            if not dataset in connCache:
                con, cur = openArticleDb(dataset)
                if con is None and preferLocal:
                    raise Exception("Could not find a local copy of Medline. Use a command line option to switch to remote NCBI Eutils lookups. For low-volume crawls (< 10000), remote lookups are sufficient.")
                connCache[dataset] = con, cur
            else:
                con, cur = connCache[dataset]

        if con is None:
            # we don't have a local database, use eutils
            art = pubPubmed.getOnePmid(pmid)
        else:
            # we have a local medline, lookup article locally
            art = lookupArticle(con, cur, "pmid", pmid)

        if art!=None:
            return art
    return None

def lookupArticle(con, cur, column, val):
    " uses sqlite db, returns a dict with info we have locally about last matching article, None if not found "
    whereExpr = "%s=%s" % (column, val)
    return list(iterArticlesWhere(con, cur, whereExpr))[-1]

def iterArticlesWhere(con, cur, whereExpr):
    " yields dicts for article that satisfy where expression "
    rows = None
    tryCount = 60

    while rows==None and tryCount>0:
        try:
            rows = list(cur.execute("SELECT * from articles WHERE %s" % whereExpr))
        except sqlite3.OperationalError:
            logging.info("Database is locked, waiting for 60 secs")
            time.sleep(60)
            tryCount -= 1

    if rows == None:
        raise Exception("database was locked for more than 60 minutes")

    if len(rows)==0:
        logging.warn("No info in local db for %s" % (whereExpr))
    # the last entry should be the newest one

    for row in rows:
        # convert sqlite object to normal dict with strings
        rowDict = {}
        for key, val in zip(list(row.keys()), row):
            rowDict[key] = val
        yield rowDict

def lookupArticleData(articleId, lookupKey="articleId"):
    " lookup article meta data for an article via a database "
    #conn = maxTables.hgSqlConnect(pubConf.mysqlDb, charset="utf8", use_unicode=True)
    #sql = "SELECT * from %s where articleId=%s" % (dataset, articleId)
    #rows = maxTables.sqlGetRows(conn,sql)
    if lookupKey=="pmid":
        dataset = "medline"
    elif lookupKey=="articleId":
        dataset = articleIdToDataset(articleId)
    assert(dataset!=None)
    textDir = join(pubConf.textBaseDir, dataset)

    if textDir not in conCache:
        dbPath = getArtDbPath(textDir)
        #assert(isfile(dbPath))
        if not (isfile(dbPath)):
            return None
        cur, con = maxTables.openSqlite(dbPath, asDict=True)
        conCache[textDir] = (cur, con)
    else:
        cur, con = conCache[textDir]

    sql = "SELECT * from articles where %s=%s" % (lookupKey, articleId)
    rows = list(cur.execute(sql))
    #assert(len(rows)==1)
    if len(rows)==0:
        #raise Exception("Could not find article %s in textDir %s" % (articleId, textDir))
        logging.error("Could not find article %s in textDir %s" % (articleId, textDir))
        return None
    articleData = rows[0]
    #authors = row["authors"]
    #author = author.split(",")[0]+" et al., "+row["journal"]
    #title = row["title"]
    #year = row["year"]
    #journal = row["journal"]
    #title = title.encode("latin1").decode("utf8")
    #text = '<small>%s (%s)</small><br><a href="%s">%s</a>' % (author, dataset, row["fulltextUrl"], title)
    return articleData

def makeChunkPath(inDir, chunkId, tabType="articles"):
    " construct path to articles table file given input dir and chunkId "
    artPath = join(inDir, "%s.%s" % (chunkId, tabType))
    if not isfile(artPath):
        artPath = join(inDir, "%s.%s.gz" % (chunkId, tabType))
    return artPath

def lookupFullDocs(inDirs, whereExpr):
    """
    inDirs is a list of directories with pubStore sqlite indexes (article.db files)
    query is the part behind the WHERE in an sql command on the articles table
    yields tuples of (articleRow, list of fileRow).
    """
    for inDir in inDirs:
        con, cur = openArticleDb(inDir, mustOpen=True)
        fileRows = []
        for artDict in iterArticlesWhere(con, cur, whereExpr):
            artId = artDict["articleId"]
            chunkId, offset = artDict["chunkId"], artDict["offset"]
            filesPath = makeChunkPath(inDir, chunkId, "files")
            logging.debug("Opening %s" % filesPath)
            filesFh = Bio.bgzf.open(filesPath)
            fileReader = maxCommon.TsvReader(filesFh)
            logging.debug("Seeking to offset %s" % offset)
            fileReader.seek(int(offset))

            # pull out all files with this article ID
            while True:
                row = fileReader.nextRow()
                if row is None or row.fileId[:len(artId)]!=artId:
                    break
                fileRows.append(row._asdict())
            yield artDict, fileRows

def getAllArticleFnames(inDir, type="articles"):
    " find all article chunks in inDir "
    inFnames = glob.glob(join(inDir, "*.%s.gz" % type))
    if len(inFnames)==0:
        inFnames = glob.glob(join(inDir, "*.%s" % type))
    return inFnames


def iterPubReaders(inDir):
    " yield a PubReader for each input chunk in inDir "
    fnames = getAllArticleFnames(inDir)
    pm = maxCommon.ProgressMeter(len(fnames))
    logging.info("found %d input chunks in %s" % (len(fnames), inDir))
    for fname in fnames:
        pr = PubReaderFile(fname)
        yield pr
        pm.taskCompleted()

def dictToMarkLines(d):
    """ convert a dict to newline-sep strings like '||key: value' and return long string """
    lines = []
    keys = list(d.keys())
    for key in sorted(keys):
        val = d[key]
        if type(val)==int:
            val = str(val)
        if val=="":
            continue
        if key!="content":
            lines.append("|%s: %s" % (key.encode("utf8"), val.encode('utf')))

    contStr = d.get("content")
    if contStr is not None:
        lines.append(d["content"].replace("\a", "\n").encode("utf8"))

    return "\n".join(lines)

if __name__=="__main__":
    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-

"""
Sci-API Unofficial API
[Search|Download] research papers from [scholar.google.com|sci-hub.io].

@author zaytoun
"""

import os
import re
import logging
import hashlib
import argparse
import requests

from bs4 import BeautifulSoup

# for now, suppress warnings due to unverified HTTPS request; this will be fixed
from requests.packages.urllib3.exceptions import InsecureRequestWarning
requests.packages.urllib3.disable_warnings(InsecureRequestWarning)

# constants
SCIHUB_BASE_URL = 'http://sci-hub.cc/'
SCHOLARS_BASE_URL = 'https://scholar.google.com/scholar'
HEADERS = {'User-Agent': 'Mozilla/5.0 (X11; Linux x86_64; rv:27.0) Gecko/20111101 Firefox/27.0'}

class SciHub(object):
    """
    SciHub class can search for papers on Google Scholars 
    and fetch/download papers from sci-hub.io
    """
    def __init__(self):
        pass

    def search(self, query, limit=10, download=False):
        """
        Performs a query on scholar.google.com, and returns a dictionary
        of results in the form {'papers': ...}. Unfortunately, as of now,
        captchas can potentially prevent searches after a certain limit.
        """
        start = 0
        results = {'papers': []}

        while True:
            try:
                res = requests.get(SCHOLARS_BASE_URL, params={'q': query, 'start': start}, headers=HEADERS)
            except requests.exceptions.RequestException as e:
                results['err'] = 'Failed to complete search with query %s (connection error)' % query
                return results

            s = self._get_soup(res.content)
            papers = s.find_all('div', class_="gs_r")
            
            if not papers:
                if 'CaptchaRedirect' in res.content:
                    results['err'] = 'Failed to complete search with query %s (captcha)' % query
                return results
            
            for paper in papers:
                if not paper.find('table'):
                    source = None
                    pdf = paper.find('div', class_='gs_ggs gs_fl')
                    link = paper.find('h3', class_='gs_rt')
                    
                    if pdf:
                        source = pdf.find('div', class_='gs_md_wp gs_ttss').find('a')['href']
                    elif link.find('a'):
                        source = link.find('a')['href']
                    else:
                        continue
                    
                    results['papers'].append({
                        'name': link.text,
                        'url': source
                    })
                    
                    if len(results['papers']) >= limit:
                        return results
            
            start += 10

    def download(self, identifier, destination='', path=None):
        """
        Downloads a paper from sci-hub given an indentifier (DOI, PMID, URL).
        Currently, this can potentially be blocked by a captcha if a certain
        limit has been reached.
        """
        data = self.fetch(identifier)

        if not 'err' in data:
            self._save(data['pdf'], 
                       os.path.join(destination, path if path else data['name']))
        
        return data

    def fetch(self, identifier):
        """
        Fetches the paper by first retrieving the direct link to the pdf.
        If the indentifier is a DOI, PMID, or URL pay-wall, then use Sci-Hub
        to access and download paper. Otherwise, just download paper directly.
        """
        url = self._get_direct_url(identifier)

        try:
            # verify=False is dangerous but sci-hub.io 
            # requires intermediate certificates to verify
            # and requests doesn't know how to download them.
            # as a hacky fix, you can add them to your store
            # and verifying would work. will fix this later.
            res = requests.get(url, headers=HEADERS, verify=False)
            return {
                'pdf': res.content,
                'url': url,
                'name': self._generate_name(res)
            }
        except requests.exceptions.RequestException as e:
            return {
                'err': 'Failed to fetch pdf with identifier %s (resolved url %s) due to %s' 
                   % (identifier, url, 'failed connection' if url else 'captcha')
            }

    def _get_direct_url(self, identifier):
        """
        Finds the direct source url for a given identifier.
        """
        id_type = self._classify(identifier)

        return identifier if id_type == 'url-direct' \
                else self._search_direct_url(identifier)

    def _search_direct_url(self, identifier):
        """
        Sci-Hub embeds papers in an iframe. This function finds the actual
        source url which looks something like https://moscow.sci-hub.io/.../....pdf.
        """
        res = requests.get(SCIHUB_BASE_URL + identifier, headers=HEADERS, verify=False)
        s = self._get_soup(res.content)
        iframe = s.find('iframe')
        if iframe:
            return iframe.get('src')

    def _classify(self, identifier):
        """
        Classify the type of identifier:
        url-direct - openly accessible paper
        url-non-direct - pay-walled paper
        pmid - PubMed ID
        doi - digital object identifier
        """
        if (identifier.startswith('http') or identifier.startswith('https')):
            if identifier.endswith('pdf'):
                return 'url-direct'
            else:
                return 'url-non-direct'
        elif identifier.isdigit():
            return 'pmid'
        else:
            return 'doi'

    def _save(self, data, path):
        """
        Save a file give data and a path.
        """
        with open(path, 'wb') as f:
            f.write(data)

    def _get_soup(self, html):
        """
        Return html soup.
        """
        return BeautifulSoup(html, 'html.parser')

    def _generate_name(self, res):
        """
        Generate unique filename for paper. Returns a name by calcuating 
        md5 hash of file contents, then appending the last 20 characters
        of the url which typically provides a good paper identifier.
        """
        name = res.url.split('/')[-1]
        pdf_hash = hashlib.md5(res.content).hexdigest()
        return '%s-%s' % (pdf_hash, name[-20:])

def main():
    sh = SciHub()

    logging.basicConfig()
    logger = logging.getLogger('Sci-Hub')

    parser = argparse.ArgumentParser(description='SciHub - To remove all barriers in the way of science.')
    parser.add_argument('-d', '--download', metavar='(DOI|PMID|URL)', help='tries to find and download the paper', type=str)
    parser.add_argument('-f', '--file', metavar='path', help='pass file with list of identifiers and download each', type=str)
    parser.add_argument('-s', '--search', metavar='query', help='search Google Scholars', type=str)
    parser.add_argument('-sd', '--search_download', metavar='query', help='search Google Scholars and download if possible', type=str)
    parser.add_argument('-l', '--limit', metavar='N', help='the number of search results to limit to', default=10, type=int)
    parser.add_argument('-o', '--output', metavar='path', help='directory to store papers', default='', type=str)
    parser.add_argument('-v', '--verbose', help='increase output verbosity', action='store_true')
    
    args = parser.parse_args()

    if args.verbose:
        logger.setLevel(logging.DEBUG)

    if args.download:
        result = sh.download(args.download, args.output)
        if 'err' in result:
            logger.debug('%s', result['err'])
        else:
            logger.debug('Successfully downloaded file with identifier %s', args.download)
    elif args.search:
        results = sh.search(args.search, args.limit)
        if 'err' in results:
            logger.debug('%s', results['err'])
        else:
            logger.debug('Successfully completed search with query %s', args.search)
        print(results)
    elif args.search_download:
        results = sh.search(args.search, args.limit)
        if 'err' in results:
            logger.debug('%s', results['err'])
        else:
            logger.debug('Successfully completed search with query %s', args.search)
            for paper in results['papers']:
                result = sh.download(paper['url'], args.output)
                if 'err' in result:
                    logger.debug('%s', result['err'])
                else:
                    logger.debug('Successfully downloaded file with identifier %s', paper['url'])
    elif args.file:
        with open(args.file, 'r') as f:
            identifiers = f.read().splitlines()
            for identifier in identifiers:
                result = sh.download(identifier, args.output)
                if 'err' in result:
                    logger.debug('%s', result['err'])
                else:
                    logger.debug('Successfully downloaded file with identifier %s', identifier)

if __name__ == '__main__':
    main()
